#!/usr/bin/env python

"""Tests for `magicgui` package."""

import inspect
from enum import Enum
from typing import NewType, Optional, Union
from unittest.mock import Mock

import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QScrollArea, QWidget

from magicgui import magicgui, register_type, type_map, widgets
from magicgui.signature import MagicSignature, magic_signature


def func(a: str = "works", b: int = 3, c=7.1) -> str:
    return a + str(b)


@pytest.fixture
def magic_func():
    """Test function decorated by magicgui."""
    return magicgui(func, call_button="my_button", auto_call=True, labels=False)


@pytest.fixture
def magic_func_defaults():
    return magicgui(func)


@pytest.fixture
def magic_func_autocall():
    return magicgui(func, auto_call=True)


def test_magicgui(magic_func):
    """Test basic magicgui functionality."""
    assert magic_func() == "works3"
    assert magic_func.a.value == "works"
    assert magic_func.b.value == 3
    assert magic_func.c.value == 7.1
    assert isinstance(magic_func.a, widgets.LineEdit)
    assert isinstance(magic_func.b, widgets.SpinBox)
    assert isinstance(magic_func.c, widgets.FloatSpinBox)

    magic_func.show()
    assert magic_func.visible

    a = magic_func.a  # save ref
    assert magic_func.index(a) == 0
    # we can delete widgets
    del magic_func.a
    with pytest.raises(AttributeError):
        _ = magic_func.a

    # they disappear from the layout
    with pytest.raises(ValueError):
        magic_func.index(a)


def test_default_call_button_behavior(magic_func_defaults, magic_func_autocall):
    assert magic_func_defaults.call_button is not None

    assert magic_func_autocall.call_button is None
    prior_autocall_count = magic_func_autocall.call_count
    magic_func_autocall.a.value = "hello"
    magic_func_autocall.b.value = 7
    assert magic_func_autocall.call_count == prior_autocall_count + 2


def test_overriding_widget_type():
    """Test overriding the widget type of a parameter."""

    # a will now be a LineEdit instead of a spinbox
    @magicgui(a={"widget_type": "LineEdit"})
    def func(a: int = 1):
        pass

    assert isinstance(func.a, widgets.LineEdit)
    assert func.a.value == "1"

    # also without type annotation
    @magicgui(a={"widget_type": "LogSlider"})
    def g(a): ...

    assert isinstance(g.a, widgets.LogSlider)


def test_unrecognized_types():
    """Test that arg with an unrecognized type is hidden."""

    class Something:
        pass

    # don't know how to handle Something type
    @magicgui
    def func(arg: Something, b: int = 1):
        pass

    assert isinstance(func.arg, widgets.EmptyWidget)

    with pytest.raises(TypeError) as e:
        func()
    assert "missing a required argument" in str(e)


def test_no_type_provided():
    """Test position args with unknown type."""

    @magicgui
    def func(a):
        pass

    assert isinstance(func.a, widgets.EmptyWidget)
    with pytest.raises(TypeError) as e:
        func()
    assert "missing a required argument" in str(e)
    assert "@magicgui(a={'bind': value})" in str(e)


def test_bind_out_of_order():
    """Test that binding a value before a non-default argument still gives message."""

    @magicgui(a={"bind": 10})
    def func(a, x):
        pass

    assert isinstance(func.a, widgets.EmptyWidget)
    with pytest.raises(TypeError) as e:
        func()
    assert "missing a required argument" in str(e)
    assert "@magicgui(x={'bind': value})" in str(e)


def test_call_button():
    """Test that the call button has been added, and pressing it calls the function."""

    @magicgui(call_button="my_button", auto_call=True)
    def func(a: int, b: int = 3, c=7.1):
        assert a == 7

    assert hasattr(func, "call_button")
    assert isinstance(func.call_button, widgets.PushButton)
    func.a.value = 7


@pytest.mark.filterwarnings("ignore")
def test_auto_call(qtbot, magic_func):
    """Test that changing a parameter calls the function."""
    from qtpy.QtTest import QTest

    # TODO: remove qtbot requirement so we can test other backends eventually.
    # changing the widget parameter calls the function
    with qtbot.waitSignal(magic_func.called, timeout=1000):
        magic_func.b.value = 6

    # changing the gui calls the function
    with qtbot.waitSignal(magic_func.called, timeout=1000):
        QTest.keyClick(magic_func.a.native, Qt.Key_A, Qt.ControlModifier)
        QTest.keyClick(magic_func.a.native, Qt.Key_Delete)


def test_dropdown_list_from_enum():
    """Test that enums properly populate the dropdown menu with options."""

    class Medium(Enum):
        Glass = 1.520
        Oil = 1.515
        Water = 1.333
        Air = 1.0003

    @magicgui
    def func(arg: Medium = Medium.Water): ...

    assert func.arg.value == Medium.Water
    assert isinstance(func.arg, widgets.ComboBox)
    assert list(func.arg.choices) == list(Medium.__members__.values())


def test_dropdown_list_from_choices():
    """Test that providing the 'choices' argument with a list of strings works."""
    CHOICES = ["Oil", "Water", "Air"]

    @magicgui(arg={"choices": CHOICES})
    def func(arg="Water"): ...

    assert func.arg.value == "Water"
    assert isinstance(func.arg, widgets.ComboBox)
    assert list(func.arg.choices) == CHOICES

    with pytest.raises(ValueError):
        # the default value must be in the list
        @magicgui(arg={"choices": ["Oil", "Water", "Air"]})
        def func(arg="Silicone"): ...


def test_dropdown_list_from_callable():
    """Test that providing the 'choices' argument with a callable works."""
    CHOICES = ["Oil", "Water", "Air"]

    def get_choices(gui):
        return CHOICES

    @magicgui(arg={"choices": get_choices})
    def func(arg="Water"): ...

    assert func.arg.value == "Water"
    assert isinstance(func.arg, widgets.ComboBox)
    assert list(func.arg.choices) == CHOICES

    func.reset_choices()


def test_changing_widget_attr_fails(magic_func):
    """Test set_widget will either update or change an existing widget."""
    assert magic_func.a.value == "works"
    widget1 = magic_func.a
    assert isinstance(widget1, widgets.LineEdit)

    # changing it to a different type will destroy and create a new widget
    widget2 = widgets.create_widget(value=1, name="a")
    with pytest.raises(AttributeError):
        magic_func.a = widget2

    assert magic_func.a == widget1


def test_multiple_gui_with_same_args():
    """Test that similarly named arguments are independent of one another."""

    @magicgui
    def example1(a=2):
        return a

    @magicgui
    def example2(a=5):
        return a

    # they get their initial values from the function sigs
    assert example1.a.value == 2
    assert example2.a.value == 5
    # settings one doesn't affect the other
    example1.a.value = 10
    assert example1.a.value == 10
    assert example2.a.value == 5
    # vice versa...
    example2.a.value = 4
    assert example1.a.value == 10
    assert example2.a.value == 4
    # calling the original equations updates the function defaults
    assert example1() == 10
    assert example2() == 4


def test_multiple_gui_instance_independence():
    """Test that multiple instance of the same decorated function are independent."""

    def example(a=2):
        return a

    w1 = magicgui(example)
    w2 = magicgui(example)
    # they get their initial values from the function sigs
    assert w1.a.value == 2
    assert w2.a.value == 2
    # settings one doesn't affect the other
    w1.a.value = 10
    assert w1.a.value == 10
    assert w2.a.value == 2
    # vice versa...
    w2.a.value = 4
    assert w1.a.value == 10
    assert w2.a.value == 4

    # all instances are independent
    assert example() == 2
    assert w1() == 10
    assert w2() == 4


def test_invisible_param():
    """Test that the visible option works."""

    @magicgui(a={"visible": False})
    def func(a: str = "string", b: int = 3, c=7.1) -> str:
        return "works"

    assert hasattr(func, "a")
    func.show()
    assert not func.a.visible
    assert func.b.visible
    assert func.c.visible
    func()


def test_bad_options():
    """Test that invalid parameter options raise TypeError."""
    with pytest.raises(TypeError):

        @magicgui(b=7)  # type: ignore
        def func(a="string", b=3, c=7.1):
            return "works"


# @pytest.mark.xfail(reason="MagicSignatures are slightly different")
def test_signature_repr():
    """Test that the gui makes a proper signature."""

    def func(a: str = "string", b: int = 3, c: float = 7.1):
        return locals()

    magic_func = magicgui(func)

    # the STRING signature representation should be the same as the original function
    assert str(inspect.signature(magic_func)) == str(inspect.signature(func))
    # however, the magic_func signature is an enhance MagicSignature object:
    assert isinstance(inspect.signature(magic_func), MagicSignature)
    assert isinstance(inspect.signature(func), inspect.Signature)

    # make sure it is up to date
    magic_func.b.value = 0
    assert (
        str(inspect.signature(magic_func))
        == "(a: str = 'string', b: int = 0, c: float = 7.1)"
    )


def test_set_choices_raises():
    """Test failures on setting choices."""

    @magicgui(mood={"choices": ["happy", "sad"]})
    def func(mood: str = "happy"):
        pass

    with pytest.raises(TypeError):
        func.mood.choices = None
    with pytest.raises(TypeError):
        func.mood.choices = 1


def test_get_choices_raises():
    """Test failures on getting choices."""

    @magicgui(mood={"choices": [1, 2, 3]})
    def func(mood: int = 1, hi: str = "hello"):
        pass

    with pytest.raises(AttributeError):
        _ = func.hi.choices

    assert func.mood.choices == (1, 2, 3)


@pytest.mark.parametrize(
    "labels",
    [
        pytest.param(
            True, marks=pytest.mark.xfail(reason="indexing still wrong with labels")
        ),
        False,
    ],
    ids=["with-labels", "no-labels"],
)
def test_add_at_position(labels):
    """Test that adding widget with position option puts widget in the right place."""

    def func(a=1, b=2, c=3):
        pass

    def get_layout_items(gui):
        lay = gui.native.layout()
        items = [lay.itemAt(i).widget()._magic_widget.name for i in range(lay.count())]
        if labels:
            items = list(filter(None, items))
        return items

    gui = magicgui(func, labels=labels)
    assert get_layout_items(gui) == ["a", "b", "c", "call_button"]
    gui.insert(1, widgets.create_widget(name="new", raise_on_unknown=False))
    assert get_layout_items(gui) == ["a", "new", "b", "c", "call_button"]


def test_original_function_works(magic_func):
    """Test that the decorated function is still operational."""
    assert magic_func() == "works3"
    assert magic_func("hi") == "hi3"


def test_show(magic_func):
    """Test that the show option works."""
    # assert not magic_func.visible
    magic_func.show()
    assert magic_func.visible


def test_register_types_by_string():
    """Test that we can register custom widget classes for certain types."""
    # must provide a non-None choices or widget_type
    with pytest.raises(ValueError):
        register_type(str, choices=None)

    register_type(int, widget_type="LineEdit")

    # this works, but choices overrides widget_type, and warns the user
    with pytest.warns(UserWarning):
        register_type(str, choices=["works", "cool", "huh"], widget_type="LineEdit")

    class Main:
        pass

    class Sub(Main):
        pass

    class Main2:
        pass

    class Sub2(Main2):
        pass

    register_type(Main, choices=[None, 1, 2, 3])
    register_type(Main2, widget_type="LineEdit")

    @magicgui
    def func(a: str = "works", b: int = 3, c: Sub = None, d: Sub2 = None):
        return a

    assert isinstance(func.a, widgets.ComboBox)
    assert isinstance(func.b, widgets.LineEdit)
    assert isinstance(func.c, widgets.ComboBox)
    assert isinstance(func.d, widgets.LineEdit)

    del type_map._type_map._TYPE_DEFS[str]
    del type_map._type_map._TYPE_DEFS[int]


def test_register_types_by_class():
    class MyLineEdit(widgets.LineEdit):
        pass

    class MyStr:
        pass

    register_type(MyStr, widget_type=MyLineEdit)
    w = widgets.create_widget(value=MyStr())
    assert isinstance(w, MyLineEdit)


def test_register_return_callback():
    """Test that registering a return callback works."""

    def check_value(gui, value, rettype):
        assert value == 1

    class Base:
        pass

    class Sub(Base):
        pass

    register_type(int, return_callback=check_value)
    register_type(Base, return_callback=check_value)

    try:

        @magicgui
        def func(a=1) -> int:
            return a

        func()
        with pytest.raises(AssertionError):
            func(3)

        @magicgui
        def func2(a=1) -> Sub:
            return a

        func2()
    finally:
        from magicgui.type_map._type_map import _RETURN_CALLBACKS

        _RETURN_CALLBACKS.pop(int)
        _RETURN_CALLBACKS.pop(Base)


def test_parent_changed(qtbot, magic_func: widgets.FunctionGui) -> None:
    """Test that setting a backend parent emits a signal."""
    wdg = QWidget()
    qtbot.addWidget(wdg)

    mock = Mock()
    magic_func.native_parent_changed.connect(mock)
    with qtbot.waitSignal(magic_func.native_parent_changed, timeout=1000):
        magic_func.native.setParent(wdg)

    # the backend parent is emitted
    mock.assert_called_once_with(wdg)

    with qtbot.waitSignal(magic_func.native_parent_changed, timeout=1000):
        magic_func.native.setParent(None)

    mock.assert_called_with(None)

    with pytest.warns(FutureWarning, match="'parent_changed' signal has been renamed"):
        magic_func.parent_changed  # noqa: B018


def test_function_binding():
    class MyObject:
        def __init__(self, name):
            self.name = name
            self.counter = 0.0

        @magicgui(call_button="callme", sigma={"max": 365})
        def method(self, sigma: float = 1):
            self.counter = self.counter + sigma
            return self.name, self.counter

    a = MyObject("a")
    b = MyObject("b")

    assert a.method.call_button.text == "callme"  # type: ignore
    assert a.method.sigma.max == 365
    assert a.method() == ("a", 1)
    assert b.method(sigma=4) == ("b", 4)
    assert a.method() == ("a", 2)
    assert b.method() == ("b", 5)


def test_function_binding_multiple():
    class MyObject:
        def __init__(self):
            pass

        @magicgui
        def method_0(self, sigma: float = 1):
            pass

        @magicgui
        def method_1(self, sigma: float = 2):
            pass

    a = MyObject()
    assert MyObject.method_0 is not a.method_0
    assert a.method_0 is not a.method_1
    assert a.method_0.sigma.value == 1
    assert a.method_1.sigma.value == 2


def test_call_count():
    """Test that a function gui remembers how many times it's been called."""

    @magicgui
    def func():
        pass

    assert func.call_count == 0
    func()
    func()
    assert func.call_count == 2
    func.reset_call_count()
    assert func.call_count == 0


def test_tooltips_from_numpydoc():
    """Test that numpydocs docstrings can be used for tooltips."""

    x_tooltip = "override tooltip"
    y_docstring = """A greeting, by default 'hi'. Notice how we miraculously pull
the entirety of the docstring just like that"""

    @magicgui(x={"tooltip": x_tooltip}, z={"tooltip": None})
    def func(x: int, y: str = "hi", z=None):
        """Do a little thing.

        Parameters
        ----------
        x : int
            An integer for you to use
        y : str, optional
            A greeting, by default 'hi'. Notice how we miraculously pull
            the entirety of the docstring just like that
        z : Any, optional
            No tooltip for me please.
        """

    assert func.x.tooltip == x_tooltip
    assert func.y.tooltip == y_docstring
    assert not func.z.tooltip


def test_bad_param_name_in_docstring():
    @magicgui
    def func(x: int):
        """Do a little thing.

        Parameters
        ----------
        not_x: int
            DESCRIPTION.
        """
        return x

    assert not func.x.tooltip


def test_duplicated_and_missing_params_from_numpydoc():
    """Test that numpydocs docstrings can be used for tooltips."""

    @magicgui
    def func(x, y, z=None):
        """Do a little thing.

        Parameters
        ----------
        x, y : int
            Integers for you to use
        """

    assert func.x.tooltip == "Integers for you to use"
    assert func.y.tooltip == "Integers for you to use"
    assert not func.z.tooltip


def test_tooltips_from_google_doc():
    """Test that google docstrings can be used for tooltips."""

    x_docstring = "An integer for you to use"
    y_docstring = """A greeting. Notice how we miraculously pull
the entirety of the docstring just like that"""

    @magicgui
    def func(x: int, y: str = "hi"):
        """Do a little thing.

        Args:
            x (int): An integer for you to use
            y (str, optional): A greeting. Notice how we miraculously pull
                               the entirety of the docstring just like that
        """

    assert func.x.tooltip == x_docstring
    assert func.y.tooltip == y_docstring


def test_tooltips_from_rest_doc():
    """Test that google docstrings can be used for tooltips."""

    x_docstring = "An integer for you to use"
    y_docstring = """A greeting, by default 'hi'. Notice how we miraculously pull
the entirety of the docstring just like that"""

    @magicgui
    def func(x: int, y: str = "hi", z=None):
        """Do a little thing.

        :param x: An integer for you to use
        :param y: A greeting, by default 'hi'. Notice how we miraculously pull
                  the entirety of the docstring just like that
        :type x: int
        :type y: str
        """

    assert func.x.tooltip == x_docstring
    assert func.y.tooltip == y_docstring


def test_no_tooltips_from_numpydoc():
    """Test that ``tooltips=False`` hides all tooltips."""

    @magicgui(tooltips=False)
    def func(x: int, y: str = "hi"):
        """Do a little thing.

        Parameters
        ----------
        x : int
            An integer for you to use
        y : str, optional
            A greeting, by default 'hi'
        """

    assert not func.x.tooltip
    assert not func.y.tooltip


def test_only_some_tooltips_from_numpydoc():
    """Test that we can still show some tooltips with ``tooltips=False``."""

    # tooltips=False, means docstrings won't be parsed at all, but tooltips
    # can still be manually provided.
    @magicgui(tooltips=False, y={"tooltip": "Still want a tooltip"})
    def func(x: int, y: str = "hi"):
        """Do a little thing.

        Parameters
        ----------
        x : int
            An integer for you to use
        y : str, optional
            A greeting, by default 'hi'
        """

    assert not func.x.tooltip
    assert func.y.tooltip == "Still want a tooltip"


def test_magicgui_type_error():
    with pytest.raises(TypeError):
        magicgui("not a function")  # type: ignore


def self_referencing_function(x: int = 1):
    """Function that refers to itself, and wants the FunctionGui instance."""
    return self_referencing_function


def test_magicgui_self_reference():
    """Test that self-referential magicguis work in global scopes."""
    global self_referencing_function
    f = magicgui(self_referencing_function)
    assert isinstance(f(), widgets.FunctionGui)
    assert f() is f


def test_local_magicgui_self_reference():
    """Test that self-referential magicguis work in local scopes."""

    @magicgui
    def local_self_referencing_function(x: int = 1):
        """Function that refers to itself, and wants the FunctionGui instance."""
        return local_self_referencing_function

    assert isinstance(local_self_referencing_function(), widgets.FunctionGui)


def test_empty_function():
    """Test that a function with no params works."""

    @magicgui(call_button=True)
    def f(): ...

    f.show()


def test_boolean_label():
    """Test that label can be used to set the text of a button widget."""

    @magicgui(check={"label": "ABC"})
    def test(check: bool, x=1):
        pass

    assert test.check.text == "ABC"

    with pytest.warns(UserWarning) as record:

        @magicgui(check={"text": "ABC", "label": "BCD"})
        def test2(check: bool, x=1):
            pass

    assert "'text' and 'label' are synonymous for button widgets" in str(record[0])


def test_none_defaults():
    """Make sure that an unannotated parameter with default=None is ok."""
    assert widgets.create_widget(value=None, raise_on_unknown=False).value is None

    def func(arg=None):
        return 1

    assert magicgui(func)() == 1

    assert str(magic_signature(func)) == str(magicgui(func).__signature__)


def test_update_and_dict():
    @magicgui
    def test(a: int = 1, y: str = "a"): ...

    assert test.asdict() == {"a": 1, "y": "a"}

    test.update(a=10, y="b")
    assert test.asdict() == {"a": 10, "y": "b"}

    test.update({"a": 1, "y": "a"})
    assert test.asdict() == {"a": 1, "y": "a"}

    test.update([("a", 10), ("y", "b")])
    assert test.asdict() == {"a": 10, "y": "b"}


def test_update_on_call():
    @magicgui
    def test(a: int = 1, y: str = "a"): ...

    assert test.call_count == 0
    test(a=10, y="b", update_widget=True)
    assert test.a.value == 10
    assert test.y.value == "b"
    assert test.call_count == 1


def test_partial():
    from functools import partial

    def some_func(x: int, y: str) -> str:
        return y + str(x)

    wdg = magicgui(partial(some_func, 1))
    assert len(wdg) == 2  # because of the call_button
    assert isinstance(wdg.y, widgets.LineEdit)
    assert not hasattr(wdg, "x")
    assert wdg("sdf") == "sdf1"
    assert wdg._callable_name == "some_func"

    wdg2 = magicgui(partial(some_func, y="sdf"))
    assert len(wdg2) == 3  # keyword arguments don't change the partial signature
    assert isinstance(wdg2.x, widgets.SpinBox)
    assert isinstance(wdg.y, widgets.LineEdit)
    assert wdg2.y.value == "sdf"
    assert wdg2(1) == "sdf1"


def test_curry():
    import toolz as tz

    @tz.curry
    def some_func2(x: int, y: str) -> str:
        return y + str(x)

    wdg = magicgui(some_func2(1))
    assert len(wdg) == 2  # because of the call_button
    assert isinstance(wdg.y, widgets.LineEdit)
    assert not hasattr(wdg, "x")
    assert wdg("sdf") == "sdf1"
    assert wdg._callable_name == "some_func2"

    wdg2 = magicgui(some_func2(y="sdf"))
    assert len(wdg2) == 3  # keyword arguments don't change the partial signature
    assert isinstance(wdg2.x, widgets.SpinBox)
    assert isinstance(wdg.y, widgets.LineEdit)
    assert wdg2.y.value == "sdf"
    assert wdg2(1) == "sdf1"


def test_scrollable():
    @magicgui(scrollable=True)
    def test_scrollable(a: int = 1, y: str = "a"): ...

    assert test_scrollable.native is not test_scrollable.root_native_widget
    assert not isinstance(test_scrollable.native, QScrollArea)
    assert isinstance(test_scrollable.root_native_widget, QScrollArea)

    @magicgui(scrollable=False)
    def test_nonscrollable(a: int = 1, y: str = "a"): ...

    assert test_nonscrollable.native is test_nonscrollable.root_native_widget
    assert not isinstance(test_nonscrollable.native, QScrollArea)


def test_unknown_exception_magicgui():
    """Test that an unknown type is raised as a RuntimeError."""

    class A:
        pass

    with pytest.raises(ValueError, match="No widget found for type"):

        @magicgui(raise_on_unknown=True)
        def func(a: A):
            print(a)


def test_unknown_exception_create_widget():
    """Test that an unknown type is raised as a RuntimeError."""

    class A:
        pass

    with pytest.raises(ValueError, match="No widget found for type"):
        widgets.create_widget(A, raise_on_unknown=True)
    with pytest.raises(ValueError, match="No widget found for type"):
        widgets.create_widget(A)
    assert isinstance(
        widgets.create_widget(A, raise_on_unknown=False), widgets.EmptyWidget
    )


@pytest.mark.parametrize("optional", [True, False])
def test_call_union_return_type(optional: bool):
    """registering Optional[type] should imply registering"""
    mock = Mock()

    NewInt = NewType("NewInt", int)
    register_type(Optional[NewInt], return_callback=mock)

    ReturnType = Optional[NewInt] if optional else NewInt

    @magicgui
    def func_optional(a: bool) -> ReturnType:
        return NewInt(1) if a else None

    func_optional(a=True)
    mock.assert_called_once_with(func_optional, 1, ReturnType)
    mock.reset_mock()
    func_optional(a=False)
    mock.assert_called_once_with(func_optional, None, ReturnType)


@pytest.mark.parametrize("optional", [True, False])
def test_no_duplication_call(optional):
    mock = Mock()
    mock2 = Mock()

    NewInt = NewType("NewInt", int)
    register_type(Optional[NewInt], return_callback=mock)
    register_type(NewInt, return_callback=mock)
    register_type(NewInt, return_callback=mock2)
    ReturnType = Optional[NewInt] if optional else NewInt

    @magicgui
    def func() -> ReturnType:
        return NewInt(1)

    func()

    mock.assert_called_once()
    assert mock2.call_count == (not optional)


def test_no_order():
    mock = Mock()

    register_type(Union[int, None], return_callback=mock)

    @magicgui
    def func() -> Union[None, int]:
        return 1

    func()
    mock.assert_called_once()



import pytest

from magicgui import magicgui, use_app, widgets


@pytest.mark.parametrize("scrollable", [False, True])
def test_container_widget(scrollable):
    """Test basic container functionality."""
    container = widgets.Container(labels=False, scrollable=scrollable)
    labela = widgets.Label(value="hi", name="labela")
    labelb = widgets.Label(value="hi", name="labelb")
    container.append(labela)
    container.extend([labelb])
    # different ways to index
    assert container[0] == labela
    assert container["labelb"] == labelb
    assert container[:1] == [labela]
    assert container[-1] == labelb

    with pytest.raises(RuntimeError):
        container[0] = "something"

    assert container.layout == "vertical"
    with pytest.raises(NotImplementedError):
        container.layout = "horizontal"

    assert all(x in dir(container) for x in ["labela", "labelb"])

    assert container.margins
    container.margins = (8, 8, 8, 8)
    assert container.margins == (8, 8, 8, 8)

    del container[1:]
    del container[-1]
    assert not container
    container.close()


@pytest.mark.parametrize("scrollable", [False, True])
def test_container_label_widths(scrollable):
    """Test basic container functionality."""
    container = widgets.Container(layout="vertical", scrollable=scrollable)
    labela = widgets.Label(value="hi", name="labela")
    labelb = widgets.Label(value="hi", name="I have a very long label")

    def _label_width():
        measure = use_app().get_obj("get_text_width")
        return max(
            measure(w.label)
            for w in container
            if not isinstance(w, widgets.bases.ButtonWidget)
        )

    container.append(labela)
    before = _label_width()
    container.append(labelb)
    assert _label_width() > before
    container.close()


@pytest.mark.parametrize("scrollable", [False, True])
def test_labeled_widget_container(scrollable):
    """Test that _LabeledWidgets follow their children."""
    from magicgui.widgets._concrete import _LabeledWidget

    w1 = widgets.Label(value="hi", name="w1")
    w2 = widgets.Label(value="hi", name="w2")
    container = widgets.Container(
        widgets=[w1, w2], layout="vertical", scrollable=scrollable
    )
    assert w1._labeled_widget
    lw = w1._labeled_widget()
    assert isinstance(lw, _LabeledWidget)
    assert not lw.visible
    container.show()
    assert w1.visible
    assert lw.visible
    w1.hide()
    assert not w1.visible
    assert not lw.visible
    w1.label = "another label"
    assert lw._label_widget.value == "another label"
    container.close()


@pytest.mark.parametrize("scrollable", [False, True])
def test_visible_in_container(scrollable):
    """Test that visibility depends on containers."""
    w1 = widgets.Label(value="hi", name="w1")
    w2 = widgets.Label(value="hi", name="w2")
    w3 = widgets.Label(value="hi", name="w3", visible=False)
    container = widgets.Container(widgets=[w2, w3], scrollable=scrollable)
    assert not w1.visible
    assert not w2.visible
    assert not w3.visible
    assert not container.visible
    container.show()
    assert container.visible
    assert w2.visible
    assert not w3.visible
    w1.show()
    assert w1.visible
    container.close()


def test_delete_widget():
    """We can delete widgets from containers."""
    a = widgets.Label(name="a")
    container = widgets.Container(widgets=[a])
    # we can delete widgets
    del container.a
    with pytest.raises(AttributeError):
        _ = container.a

    # they disappear from the layout
    with pytest.raises(ValueError):
        container.index(a)


def test_reset_choice_recursion():
    """Test that reset_choices recursion works for multiple types of widgets."""
    x = 0

    def get_choices(widget):
        nonlocal x
        x += 1
        return list(range(x))

    @magicgui(c={"choices": get_choices})
    def f(c):
        pass

    assert f.c.choices == (0,)

    container = widgets.Container(widgets=[f])
    container.reset_choices()
    assert f.c.choices == (0, 1)
    container.reset_choices()
    assert f.c.choices == (0, 1, 2)


def test_container_indexing_with_native_mucking():
    """Mostly make sure that the inner model isn't messed up.

    keeping indexes with a manipulated native model *may* be something to do in future.
    """
    l1 = widgets.Label(name="l1")
    l2 = widgets.Label(name="l2")
    l3 = widgets.Label(name="l3")
    c = widgets.Container(widgets=[l1, l2, l3])
    assert c[-1] == l3
    # so far they should be in sync
    native = c.native.layout()
    assert native.count() == len(c)
    # much with native layout
    native.addStretch()
    # haven't changed the magicgui container
    assert len(c) == 3
    assert c[-1] == l3
    # though it has changed the native model
    assert native.count() == 4


def test_containers_show_nested_containers():
    """make sure showing a container shows a nested FunctionGui."""

    @magicgui
    def func(x: int, y: str):
        pass

    assert not func.visible
    c2 = widgets.Container(widgets=[func])
    assert not c2.visible
    c2.show()
    assert c2.visible and func.visible
    c2.close()
    assert not func.visible


def test_container_removal():
    c = widgets.Container()
    s = widgets.Slider(label="label")
    assert len(c) == 0
    assert c.native.layout().count() == 0

    c.append(s)
    assert len(c) == 1
    assert c.native.layout().count() == 1

    c.pop()
    assert len(c) == 0
    assert c.native.layout().count() == 0


def test_connection_during_init() -> None:
    class C(widgets.Container):
        def __init__(self) -> None:
            btn = widgets.PushButton()
            btn.changed.connect(self._on_clicked)
            super().__init__(widgets=[btn])

        def _on_clicked(self): ...

    assert isinstance(C(), widgets.Container)


def test_parent():
    lbl = widgets.Label()
    inner = widgets.Container(widgets=[lbl])
    outer = widgets.Container()
    outer.append(inner)

    assert lbl.parent is inner
    assert inner.parent is outer
    assert outer.parent is None



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from magicgui.widgets._concrete import backend_widget
from magicgui.widgets.bases import ValueWidget

if TYPE_CHECKING:
    from pathlib import Path

    import matplotlib.colors
    import numpy as np
    import PIL.Image

    from magicgui.widgets.protocols import ValueWidgetProtocol

    from . import _mpl_image
    from ._mpl_image import Colormap, Normalize


@backend_widget
class Image(ValueWidget):
    """A non-editable image display."""

    _widget: ValueWidgetProtocol
    _image: _mpl_image.Image | None = None

    @property
    def value(self):
        """Return current image array."""
        return self._image._A if self._image else None

    @value.setter
    def value(self, value):
        """Set current data.  Alias for ``image.set_data(value)``."""
        self.set_data(value)

    def set_data(
        self,
        val: str | Path | np.ndarray | PIL.Image.Image,
        cmap: str | Colormap | matplotlib.colors.Colormap | None = None,
        norm: _mpl_image.Normalize | matplotlib.colors.Normalize | None = None,
        vmin: float | None = None,
        vmax: float | None = None,
        width: int | Literal["auto"] | None = None,
        height: int | Literal["auto"] | None = None,
        format: str | None = None,
    ):
        """Set image data with various optional display parameters.

        Parameters
        ----------
        val : str, Path, np.ndarray or PIL.Image.Image
            The image data or file to load. Data must be 2D (monochromatic), or
            3D: MxNx3 (RGB) or MxNx4 (RGBA).
        cmap : str, magicgui.types.Colormap, or matplotlib.colors.Colormap, optional
            A colormap to use for monochromatic images.  If a string, matplotlib must
            be installed and the colormap will be selected with ``cm.get_cmap(cmap)``.
        norm : magicgui.types.Normalize, or matplotlib.colors.Normalize, optional
            A normalization object to use for rendering images.  Accepts matplotlib
            Normalize objects.
        vmin : float, optional
            The min contrast limit to use when scaling monochromatic images
        vmax : float, optional
            The max contrast limit to use when scaling monochromatic images
        width : int or "auto", optional
            Set the width of the widget. If "auto", sets the widget size to the image
            size (1:1). If width is provided, height is auto-set based on aspect ratio.
        height : int or "auto", optional
            Set the height of the widget. If "auto", sets the widget size to the image
            size (1:1).  If width is provided, height is auto-set based on aspect ratio.
        format : str, optional
            Force image format type for ``imread`` when ``val`` is provided as a string,
            by default None

        Raises
        ------
        TypeError
            If the provided data shape or type is invalid.
        ImportError
            If a string is provided for ``val`` and PIL is not installed.
        RuntimeError
            If a ``PIL.Image.Image`` instance is provided as data, with an unrecognized
            image mode.
        """
        if self._image is None:
            from magicgui.widgets._image import _mpl_image

            self._image = _mpl_image.Image()

        self._image.set_data(val, format=format)
        self._image.set_clim(vmin, vmax)
        self._image.set_cmap(cmap)
        self._image.set_norm(norm)

        im = self._image.make_image()
        im_height, im_width, *_ = im.shape
        if "auto" in (height, width):
            self.scale_widget_to_image_size()
        elif width:
            self.width = width  # type: ignore
            self.height = width * im_height / im_width
        elif height:
            self.height = height  # type: ignore
            self.width = height * im_width / im_height
        self._widget._mgui_set_value(im)

    def scale_widget_to_image_size(self):
        """Set the size of the widget to the size of the image."""
        if self._image is not None:
            im = self._image.make_image()
            self.width = im.shape[1]
            self.height = im.shape[0]

    @property
    def image_rgba(self) -> np.ndarray | None:
        """Return rendered numpy array."""
        return self._image.make_image() if self._image is not None else None

    @property
    def image_data(self) -> np.ndarray | None:
        """Return image data."""
        return self._image._A if self._image is not None else None

    def get_clim(self) -> tuple[float | None, float | None]:
        """Get contrast limits (for monochromatic images)."""
        return self._image.get_clim() if self._image is not None else (None, None)

    def set_clim(self, vmin: float | None = None, vmax: float | None = None):
        """Set contrast limits (for monochromatic images).

        Parameters
        ----------
        vmin : float, optional
            The min contrast limit to use when scaling monochromatic images
        vmax : float, optional
            The max contrast limit to use when scaling monochromatic images
        """
        if self._image is None:
            raise RuntimeError("You add data with `set_data` before setting clims")
        self._image.set_clim(vmin, vmax)
        self._widget._mgui_set_value(self._image.make_image())

    def set_cmap(self, cmap: str | Colormap | matplotlib.colors.Colormap):
        """Set colormap (for monochromatic images).

        Parameters
        ----------
        cmap : str, magicgui.types.Colormap, or matplotlib.colors.Colormap, optional
            A colormap to use for monochromatic images.  If a string, matplotlib must
            be installed and the colormap will be selected with ``cm.get_cmap(cmap)``.
        """
        if self._image is None:
            raise RuntimeError("You add data with `set_data` before setting cmaps")
        self._image.set_cmap(cmap)
        self._widget._mgui_set_value(self._image.make_image())

    def set_norm(self, norm: Normalize | matplotlib.colors.Normalize):
        """Set normalization method.

        Parameters
        ----------
        norm : magicgui.types.Normalize, or matplotlib.colors.Normalize, optional
            A normalization object to use for rendering images.  Accepts matplotlib
            Normalize objects.
        """
        if self._image is None:
            raise RuntimeError("You add data with `set_data` before setting norm")
        self._image.set_norm(norm)
        self._widget._mgui_set_value(self._image.make_image())

    def __repr__(self) -> str:
        """Return representation of widget of instance."""
        d = self.image_data
        shape = "x".join(map(str, d.shape)) if d is not None else "<no data>"
        return f"{self.widget_type}({shape}, name={self.name!r})"



"""Gallery configuration for docs."""

import warnings
from pathlib import Path

import napari
import qtgallery
from mkdocs_gallery.gen_data_model import GalleryScript
from mkdocs_gallery.scrapers import figure_md_or_html, matplotlib_scraper
from qtpy.QtWidgets import QApplication

warnings.filterwarnings("ignore", category=DeprecationWarning)


def qt_window_scraper(block: tuple, script: GalleryScript) -> str:
    """Scrape screenshots from open Qt windows.

    Parameters
    ----------
    block : tuple
        A tuple containing the (label, content, line_number) of the block.
    script : GalleryScript
        Script being run

    Returns
    -------
    md : str
        The ReSTructuredText that will be rendered to HTML containing
        the images. This is often produced by :func:`figure_md_or_html`.
    """
    imgpath_iter = script.run_vars.image_path_iterator

    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    app.processEvents()

    # get top-level widgets that aren't hidden
    widgets = [w for w in app.topLevelWidgets() if not w.isHidden()]

    image_paths = []
    for widg, imgpath in zip(widgets, imgpath_iter):
        pixmap = widg.grab()
        pixmap.save(str(imgpath))
        image_paths.append(imgpath)
        widg.close()

    return figure_md_or_html(image_paths, script)


def napari_image_scraper(block, script: GalleryScript):
    """Scrape screenshots from napari windows.

    Parameters
    ----------
    block : tuple
        A tuple containing the (label, content, line_number) of the block.
    script : GalleryScript
        Script being run

    Returns
    -------
    md : str
        The ReSTructuredText that will be rendered to HTML containing
        the images. This is often produced by :func:`figure_md_or_html`.
    """
    viewer = napari.current_viewer()
    if viewer is not None:
        image_path = next(script.run_vars.image_path_iterator)
        viewer.screenshot(canvas_only=False, flash=False, path=image_path)
        viewer.close()
        return figure_md_or_html([image_path], script)
    else:
        return ""


def _reset_napari(gallery_conf, file: Path):
    # Close all open napari windows and reset theme
    while napari.current_viewer() is not None:
        napari.current_viewer().close()
    settings = napari.settings.get_settings()
    settings.appearance.theme = "dark"
    # qtgallery manages the event loop so it
    # is not completely blocked by napari.run()
    qtgallery.reset_qapp(gallery_conf, file)


conf = {
    "image_scrapers": [napari_image_scraper, qt_window_scraper, matplotlib_scraper],
    "reset_modules": [_reset_napari, qtgallery.reset_qapp],
}



"""# Indeterminate progress bar

Example of an indeterminate progress bar for a long running computation
of unknown time.

"""

import time

from superqt.utils import thread_worker

from magicgui import magicgui
from magicgui.tqdm import tqdm


@magicgui(call_button=True, layout="horizontal")
def long_running(sleep_time=5):
    """Long running computation with an indeterminate progress bar."""
    # Here tqdm is not provided an iterable argument, or the 'total' kwarg
    # so it cannot calculate the expected number of iterations
    # which means it will create an indeterminate progress bar
    with tqdm() as pbar:
        # It is best practice to use a separate thread for long running computations
        # This makes the function non-blocking, you can still interact with the widget
        @thread_worker(connect={"finished": lambda: pbar.progressbar.hide()})
        def sleep(secs):
            time.sleep(secs)

        sleep(sleep_time)


long_running.show(run=True)



"""Widget implementations (adaptors) for the Qt backend."""

from __future__ import annotations

import math
import re
import warnings
from contextlib import contextmanager, suppress
from functools import partial
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, Iterable, Iterator, Sequence

import qtpy
import superqt
from qtpy import QtWidgets as QtW
from qtpy.QtCore import QEvent, QObject, QSize, Qt, Signal
from qtpy.QtGui import (
    QFont,
    QFontMetrics,
    QIcon,
    QImage,
    QKeyEvent,
    QPalette,
    QPixmap,
    QResizeEvent,
    QTextDocument,
)

from magicgui.types import FileDialogMode, Separator
from magicgui.widgets import Widget, protocols
from magicgui.widgets._concrete import _LabeledWidget

if TYPE_CHECKING:
    import numpy


@contextmanager
def _signals_blocked(obj: QtW.QWidget) -> Iterator[None]:
    before = obj.blockSignals(True)
    try:
        yield
    finally:
        obj.blockSignals(before)


if TYPE_CHECKING:
    import numpy as np


class EventFilter(QObject):
    parentChanged = Signal(QObject)
    valueChanged = Signal(object)
    paletteChanged = Signal()

    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Type.ParentChange:
            self.parentChanged.emit(obj.parent())
        if event.type() == QEvent.Type.PaletteChange:
            self.paletteChanged.emit()
        return False


class QBaseWidget(protocols.WidgetProtocol):
    """Implements show/hide/native."""

    _qwidget: QtW.QWidget

    def __init__(
        self, qwidg: type[QtW.QWidget], parent: QtW.QWidget | None = None, **kwargs: Any
    ) -> None:
        self._qwidget = qwidg(parent=parent)
        self._qwidget.setObjectName(f"magicgui.{qwidg.__name__}")
        self._event_filter = EventFilter()
        self._qwidget.installEventFilter(self._event_filter)

    def _mgui_close_widget(self) -> None:
        self._qwidget.close()

    def _mgui_get_visible(self) -> bool:
        return self._qwidget.isVisible()

    def _mgui_set_visible(self, value: bool) -> None:
        self._qwidget.setVisible(value)

    def _mgui_get_enabled(self) -> bool:
        return self._qwidget.isEnabled()

    def _mgui_set_enabled(self, enabled: bool) -> None:
        self._qwidget.setEnabled(enabled)

    def _mgui_get_parent(self) -> Widget | None:
        par = self._qwidget.parent()
        # FIXME: This whole thing is hacky.
        while par is not None:
            mgui_wdg = getattr(par, "_magic_widget", None)
            # the labeled widget itself should be considered a "hidden" layer.
            if isinstance(mgui_wdg, Widget) and not isinstance(
                mgui_wdg, _LabeledWidget
            ):
                return mgui_wdg
            par = par.parent()
        return None

    def _mgui_set_parent(self, widget: Widget) -> None:
        self._qwidget.setParent(widget.native if widget else None)

    def _mgui_get_native_widget(self) -> QtW.QWidget:
        return self._qwidget

    def _mgui_get_root_native_widget(self) -> QtW.QWidget:
        return self._qwidget

    def _mgui_get_width(self) -> int:
        """Return the current width of the widget."""
        return self._qwidget.width()

    def _mgui_set_width(self, value: int) -> None:
        """Set the current width of the widget."""
        self._qwidget.resize(int(value), self._qwidget.height())

    def _mgui_get_min_width(self) -> int:
        """Get the minimum allowable width of the widget."""
        return self._qwidget.minimumWidth()

    def _mgui_set_min_width(self, value: int) -> None:
        """Set the minimum allowable width of the widget."""
        self._qwidget.setMinimumWidth(int(value))
        self._qwidget.resize(self._qwidget.sizeHint())

    def _mgui_get_max_width(self) -> int:
        """Get the maximum allowable width of the widget."""
        return self._qwidget.maximumWidth()

    def _mgui_set_max_width(self, value: int) -> None:
        """Set the maximum allowable width of the widget."""
        self._qwidget.setMaximumWidth(int(value))
        self._qwidget.resize(self._qwidget.sizeHint())

    def _mgui_get_height(self) -> int:
        """Return the current height of the widget."""
        return self._qwidget.height()

    def _mgui_set_height(self, value: int) -> None:
        """Set the current height of the widget."""
        self._qwidget.resize(self._qwidget.width(), int(value))

    def _mgui_get_min_height(self) -> int:
        """Get the minimum allowable height of the widget."""
        return self._qwidget.minimumHeight()

    def _mgui_set_min_height(self, value: int) -> None:
        """Set the minimum allowable height of the widget."""
        self._qwidget.setMinimumHeight(int(value))
        self._qwidget.resize(self._qwidget.sizeHint())

    def _mgui_get_max_height(self) -> int:
        """Get the maximum allowable height of the widget."""
        return self._qwidget.maximumHeight()

    def _mgui_set_max_height(self, value: int) -> None:
        """Set the maximum allowable height of the widget."""
        self._qwidget.setMaximumHeight(int(value))
        self._qwidget.resize(self._qwidget.sizeHint())

    def _mgui_get_tooltip(self) -> str:
        return self._qwidget.toolTip()

    def _mgui_set_tooltip(self, value: str | None) -> None:
        self._qwidget.setToolTip(str(value) if value else "")

    def _mgui_bind_parent_change_callback(self, callback: Callable) -> None:
        self._event_filter.parentChanged.connect(callback)

    def _mgui_render(self) -> numpy.ndarray:
        try:
            import numpy as np
        except ImportError:
            raise ModuleNotFoundError(
                "could not find module 'numpy'. "
                "Please `pip install numpy` to render widgets."
            ) from None

        img = self._qwidget.grab().toImage()
        if img.format() != QImage.Format_ARGB32:
            img = img.convertToFormat(QImage.Format_ARGB32)
        bits = img.constBits()
        h, w, c = img.height(), img.width(), 4
        if qtpy.API_NAME.startswith("PySide"):
            arr = np.array(bits).reshape(h, w, c)
        else:
            bits.setsize(h * w * c)
            arr = np.frombuffer(bits, np.uint8).reshape(h, w, c)  # type: ignore

        return arr[:, :, [2, 1, 0, 3]]


class QBaseValueWidget(QBaseWidget, protocols.ValueWidgetProtocol):
    """Implements get/set/bind_change."""

    def __init__(
        self,
        qwidg: type[QtW.QWidget],
        getter: str,
        setter: str,
        onchange: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(qwidg, **kwargs)
        self._getter_name = getter
        self._setter_name = setter
        self._onchange_name = onchange

    def _mgui_get_value(self) -> Any:
        val = getattr(self._qwidget, self._getter_name)()
        return self._post_get_hook(val)

    def _mgui_set_value(self, value: Any) -> None:
        val = self._pre_set_hook(value)
        getattr(self._qwidget, self._setter_name)(val)

    def _mgui_bind_change_callback(self, callback: Callable) -> None:
        signal_instance = getattr(self._qwidget, self._onchange_name, None)
        if signal_instance:
            signal_instance.connect(callback)

    def _post_get_hook(self, value: Any) -> Any:
        return value

    def _pre_set_hook(self, value: Any) -> Any:
        return value


# BASE WIDGET


class EmptyWidget(QBaseWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QWidget, **kwargs)

    def _mgui_get_value(self) -> Any:
        raise NotImplementedError()

    def _mgui_set_value(self, value: Any) -> None:
        raise NotImplementedError()

    def _mgui_bind_change_callback(self, callback: Callable) -> None:
        pass


# STRING WIDGETS


class QBaseStringWidget(QBaseValueWidget):
    _qwidget: QtW.QLineEdit | QtW.QTextEdit | QtW.QLabel

    def _mgui_set_value(self, value: Any) -> None:
        super()._mgui_set_value(str(value))


class Label(QBaseStringWidget):
    _qwidget: QtW.QLabel

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QLabel, "text", "setText", "", **kwargs)
        self._qwidget.setSizePolicy(
            QtW.QSizePolicy.Policy.Fixed, QtW.QSizePolicy.Policy.Fixed
        )

    def _mgui_bind_change_callback(self, callback: Callable) -> None:
        # raise NotImplementedError("QLabel has no change signal")
        pass

    def _mgui_set_value(self, value: Any) -> None:
        super()._mgui_set_value(str(value))


class _ResizeableLabel(QtW.QLabel):
    resized = Signal()

    def resizeEvent(self, a0: QResizeEvent) -> None:
        self.resized.emit()
        return super().resizeEvent(a0)


class Image(QBaseValueWidget):
    _qwidget: _ResizeableLabel

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(_ResizeableLabel, "text", "setText", "", **kwargs)
        self._qwidget.setSizePolicy(
            QtW.QSizePolicy.Policy.Ignored, QtW.QSizePolicy.Policy.Ignored
        )
        self._qwidget.resized.connect(self._rescale)
        self._pixmap: QPixmap | None = None

    def _rescale(self) -> None:
        if self._pixmap:
            sz = self._qwidget.size()
            self._qwidget.setPixmap(
                self._pixmap.scaled(
                    sz,
                    Qt.AspectRatioMode.KeepAspectRatio,
                    Qt.TransformationMode.SmoothTransformation,
                )
            )

    def _mgui_set_value(self, val: np.ndarray) -> None:
        image = QImage(val, val.shape[1], val.shape[0], QImage.Format.Format_RGBA8888)  # type: ignore
        self._pixmap = QPixmap.fromImage(image)
        self._rescale()


class QuantityEdit(QBaseValueWidget):
    _qwidget: superqt.QQuantity

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            superqt.QQuantity, "value", "setValue", "valueChanged", **kwargs
        )


class LineEdit(QBaseStringWidget):
    _qwidget: QtW.QLineEdit

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QLineEdit, "text", "setText", "textChanged", **kwargs)


class Password(LineEdit):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._qwidget.setEchoMode(QtW.QLineEdit.EchoMode.Password)


class LiteralEvalLineEdit(QBaseStringWidget):
    _qwidget: QtW.QLineEdit

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QLineEdit, "text", "setText", "textChanged", **kwargs)

    def _post_get_hook(self, value: Any) -> Any:
        from ast import literal_eval

        return literal_eval(value)


class TextEdit(QBaseStringWidget, protocols.SupportsReadOnly):
    _qwidget: QtW.QTextEdit

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            QtW.QTextEdit, "toPlainText", "setText", "textChanged", **kwargs
        )

    def _mgui_set_read_only(self, value: bool) -> None:
        self._qwidget.setReadOnly(value)

    def _mgui_get_read_only(self) -> bool:
        return self._qwidget.isReadOnly()


# NUMBERS


class QBaseRangedWidget(QBaseValueWidget, protocols.RangedWidgetProtocol):
    """Provides min/max/step implementations."""

    _qwidget: QtW.QDoubleSpinBox | superqt.QLargeIntSpinBox | QtW.QAbstractSlider
    _precision: float = 1

    def __init__(self, qwidg: type[QtW.QWidget], **kwargs: Any) -> None:
        super().__init__(qwidg, "value", "setValue", "valueChanged", **kwargs)

    def _mgui_get_min(self) -> float:
        """Get the minimum possible value."""
        val = self._qwidget.minimum()
        return self._post_get_hook(val)

    def _mgui_set_min(self, value: float):
        """Set the minimum possible value."""
        self._update_precision(minimum=value)
        val = self._pre_set_hook(value)
        self._qwidget.setMinimum(val)

    def _mgui_get_max(self) -> float:
        """Set the maximum possible value."""
        val = self._qwidget.maximum()
        return self._post_get_hook(val)

    def _mgui_set_max(self, value: float):
        """Set the maximum possible value."""
        self._update_precision(maximum=value)
        val = self._pre_set_hook(value)
        self._qwidget.setMaximum(val)

    def _mgui_get_step(self) -> float:
        """Get the step size."""
        val = self._qwidget.singleStep()
        return self._post_get_hook(val)

    def _mgui_set_adaptive_step(self, value: bool):
        """Seti is widget single steep should be adaptive."""
        self._qwidget.setStepType(
            QtW.QAbstractSpinBox.StepType.AdaptiveDecimalStepType
            if value
            else QtW.QAbstractSpinBox.StepType.DefaultStepType
        )

    def _mgui_get_adaptive_step(self) -> bool:
        """Get is widget single steep should be adaptive."""
        return (
            self._qwidget.stepType()
            == QtW.QAbstractSpinBox.StepType.AdaptiveDecimalStepType
        )

    def _mgui_set_step(self, value: float):
        """Set the step size."""
        self._update_precision(step=value)
        val = self._pre_set_hook(value)
        self._qwidget.setSingleStep(val)

    def _update_precision(self, **kwargs: Any) -> None:
        pass


# BUTTONS


class QBaseButtonWidget(
    QBaseValueWidget, protocols.SupportsText, protocols.SupportsIcon
):
    _qwidget: QtW.QCheckBox | QtW.QPushButton | QtW.QRadioButton | QtW.QToolButton

    def __init__(self, qwidg: type[QtW.QAbstractButton], **kwargs: Any) -> None:
        super().__init__(qwidg, "isChecked", "setChecked", "toggled", **kwargs)
        self._event_filter.paletteChanged.connect(self._update_icon)
        self._icon: tuple[str | None, str | None] | None = None

    def _mgui_set_text(self, value: str) -> None:
        """Set text."""
        self._qwidget.setText(str(value))

    def _mgui_get_text(self) -> str:
        """Get text."""
        return self._qwidget.text()

    def _update_icon(self) -> None:
        # Called when palette changes or icon is set
        if self._icon:
            qicon = _get_qicon(*self._icon, palette=self._qwidget.palette())
            if qicon is None:
                self._icon = None  # an error occurred don't try again
                self._qwidget.setIcon(QIcon())
            else:
                self._qwidget.setIcon(qicon)

    def _mgui_set_icon(self, value: str | None, color: str | None) -> None:
        self._icon = (value, color)
        self._update_icon()


def _get_qicon(key: str | None, color: str | None, palette: QPalette) -> QIcon | None:
    """Return a QIcon from iconify, or None if it fails."""
    if not key:
        return QIcon()

    if not color or color == "auto":
        # use foreground color
        color = palette.color(QPalette.ColorRole.WindowText).name()
        # don't use full black or white
        color = {"#000000": "#333333", "#ffffff": "#cccccc"}.get(color, color)

    if ":" not in key:
        # for parity with the other backends, assume fontawesome
        # if no prefix is given.
        key = f"fa:{key}"

    try:
        return superqt.QIconifyIcon(key, color=color)
    except (OSError, ValueError) as e:
        warnings.warn(f"Could not set iconify icon: {e}", stacklevel=2)
        return None


class PushButton(QBaseButtonWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QPushButton, **kwargs)
        self._onchange_name = "clicked"
        # make enter/return "click" the button when focused.
        self._qwidget.setAutoDefault(True)


class CheckBox(QBaseButtonWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QCheckBox, **kwargs)


class RadioButton(QBaseButtonWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QRadioButton, **kwargs)


# class ToolButton(QBaseButtonWidget):
#     def __init__(self,**kwargs: Any) -> None:
#         super().__init__(QtW.QToolButton, **kwargs)


class Container(
    QBaseWidget, protocols.ContainerProtocol, protocols.SupportsOrientation
):
    def __init__(
        self, layout="vertical", scrollable: bool = False, **kwargs: Any
    ) -> None:
        QBaseWidget.__init__(self, QtW.QWidget, **kwargs)
        if layout == "horizontal":
            self._layout: QtW.QBoxLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()
        self._qwidget.setLayout(self._layout)

        if scrollable:
            self._scroll = QtW.QScrollArea()
            # Allow widget to resize when window is larger than min widget size
            self._scroll.setWidgetResizable(True)
            if layout == "horizontal":
                horiz_policy = Qt.ScrollBarPolicy.ScrollBarAsNeeded
                vert_policy = Qt.ScrollBarPolicy.ScrollBarAlwaysOff
            else:
                horiz_policy = Qt.ScrollBarPolicy.ScrollBarAlwaysOff
                vert_policy = Qt.ScrollBarPolicy.ScrollBarAsNeeded
            self._scroll.setHorizontalScrollBarPolicy(horiz_policy)
            self._scroll.setVerticalScrollBarPolicy(vert_policy)
            self._scroll.setWidget(self._qwidget)
            self._qwidget = self._scroll

    @property
    def _is_scrollable(self) -> bool:
        return isinstance(self._qwidget, QtW.QScrollArea)

    def _mgui_get_root_native_widget(self):
        return self._qwidget

    def _mgui_get_native_widget(self):
        # Return widget with the layout set
        return self._qwidget.widget() if self._is_scrollable else self._qwidget

    def _mgui_get_visible(self):
        return self._mgui_get_native_widget().isVisible()

    def _mgui_insert_widget(self, position: int, widget: Widget):
        self._layout.insertWidget(position, widget.native)
        if self._is_scrollable:
            min_size = self._layout.totalMinimumSize()
            if isinstance(self._layout, QtW.QHBoxLayout):
                self._scroll.setMinimumHeight(min_size.height())
            else:
                self._scroll.setMinimumWidth(min_size.width() + 20)

    def _mgui_remove_widget(self, widget: Widget):
        self._layout.removeWidget(widget.native)
        widget.native.setParent(None)

    def _mgui_get_margins(self) -> tuple[int, int, int, int]:
        m = self._layout.contentsMargins()
        return m.left(), m.top(), m.right(), m.bottom()

    def _mgui_set_margins(self, margins: tuple[int, int, int, int]) -> None:
        self._layout.setContentsMargins(*margins)

    def _mgui_set_orientation(self, value) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        raise NotImplementedError(
            "Sorry, changing orientation after instantiation "
            "is not yet implemented for Qt."
        )

    def _mgui_get_orientation(self) -> str:
        """Set orientation, return either 'horizontal' or 'vertical'."""
        if isinstance(self, QtW.QHBoxLayout):
            return "horizontal"
        else:
            return "vertical"


class MainWindow(Container):
    def __init__(
        self, layout="vertical", scrollable: bool = False, **kwargs: Any
    ) -> None:
        super().__init__(layout=layout, scrollable=scrollable, **kwargs)
        self._main_window = QtW.QMainWindow()
        self._menus: dict[str, QtW.QMenu] = {}
        if scrollable:
            self._main_window.setCentralWidget(self._scroll)
        else:
            self._main_window.setCentralWidget(self._qwidget)

    def _mgui_get_visible(self):
        return self._main_window.isVisible()

    def _mgui_set_visible(self, value: bool):
        self._main_window.setVisible(value)

    def _mgui_get_native_widget(self) -> QtW.QMainWindow:
        return self._main_window

    def _mgui_create_menu_item(
        self,
        menu_name: str,
        action_name: str,
        callback: Callable | None = None,
        shortcut: str | None = None,
    ):
        menu = self._menus.setdefault(
            menu_name, self._main_window.menuBar().addMenu(f"&{menu_name}")
        )
        action = QtW.QAction(action_name, self._main_window)
        if shortcut is not None:
            action.setShortcut(shortcut)
        if callback is not None:
            action.triggered.connect(callback)
        menu.addAction(action)


class SpinBox(QBaseRangedWidget):
    def __init__(self, **kwargs: Any) -> None:
        # TODO: Consider any performance impace of this widget over a QSpinBox
        super().__init__(superqt.QLargeIntSpinBox, **kwargs)

    def _mgui_set_value(self, value) -> None:
        super()._mgui_set_value(int(value))

    def _pre_set_hook(self, value):
        return int(value)


class FloatSpinBox(QBaseRangedWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QDoubleSpinBox, **kwargs)

    def _mgui_set_value(self, value) -> None:
        super()._mgui_set_value(float(value))

    def _mgui_set_step(self, value: float):
        """Set the step size."""
        if value and value < 1 * 10 ** -self._qwidget.decimals():
            self._qwidget.setDecimals(math.ceil(abs(math.log10(value))))
        self._qwidget.setSingleStep(value)


class _Slider(QBaseRangedWidget, protocols.SupportsOrientation):
    _qwidget: QtW.QSlider

    def __init__(
        self,
        qwidg=QtW.QSlider,
        readout: bool = True,
        orientation: str = "horizontal",
        **kwargs,
    ):
        super().__init__(qwidg, **kwargs)
        self._mgui_set_orientation("horizontal")
        self._mgui_set_readout_visibility(readout)
        self._mgui_set_orientation(orientation)

    def _mgui_set_orientation(self, value) -> Any:
        """Get current value of the widget."""
        orientation = Qt.Vertical if value == "vertical" else Qt.Horizontal
        self._qwidget.setOrientation(orientation)

    def _mgui_get_orientation(self) -> Any:
        """Get current value of the widget."""
        orientation = self._qwidget.orientation()
        return "vertical" if orientation == Qt.Vertical else "horizontal"

    def _mgui_set_readout_visibility(self, value: bool):
        raise NotImplementedError()

    def _mgui_get_tracking(self) -> bool:
        # Progressbar also uses this base, but doesn't have tracking
        if hasattr(self._qwidget, "hasTracking"):
            return self._qwidget.hasTracking()
        return False

    def _mgui_set_tracking(self, value: bool) -> None:
        # Progressbar also uses this base, but doesn't have tracking
        if hasattr(self._qwidget, "setTracking"):
            self._qwidget.setTracking(value)

    def _pre_set_hook(self, value):
        return self._cast(value)

    def _cast(self, value: Any) -> Any:
        return int(value)


class Slider(_Slider):
    _qwidget: QtW.QSlider
    # TODO: Consider any performance impace of this widget over a QSpinBox
    _readout = superqt.QLargeIntSpinBox

    def __init__(self, qwidg=QtW.QSlider, **kwargs: Any) -> None:
        self._container = QtW.QWidget()
        self._readout_widget = self._readout()
        super().__init__(qwidg, **kwargs)

        no_buttons = QtW.QAbstractSpinBox.ButtonSymbols.NoButtons
        self._readout_widget.setButtonSymbols(no_buttons)
        self._readout_widget.setStyleSheet("background:transparent; border: 0;")

        self._qwidget.valueChanged.connect(self._on_slider_change)
        self._readout_widget.editingFinished.connect(self._on_readout_change)

    def _mgui_get_visible(self):
        return self._container.isVisible()

    def _mgui_set_visible(self, value: bool):
        self._container.setVisible(value)

    def _mgui_set_orientation(self, value: str) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        if value == "vertical":
            layout = QtW.QVBoxLayout()
            self._qwidget.setOrientation(Qt.Vertical)
            layout.addWidget(self._qwidget, alignment=Qt.AlignmentFlag.AlignHCenter)
            layout.addWidget(
                self._readout_widget, alignment=Qt.AlignmentFlag.AlignHCenter
            )
            self._readout_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)
            layout.setContentsMargins(0, 0, 0, 0)
            layout.setSpacing(1)
        else:
            layout = QtW.QHBoxLayout()
            self._qwidget.setOrientation(Qt.Horizontal)
            layout.addWidget(self._qwidget)
            layout.addWidget(self._readout_widget)
            self._readout_widget.setAlignment(Qt.AlignmentFlag.AlignRight)
            right_margin = 0 if self._readout_widget.isVisible() else 4
            layout.setContentsMargins(0, 0, right_margin, 0)
            layout.setSpacing(4)
        old_layout = self._container.layout()
        if old_layout is not None:
            QtW.QWidget().setLayout(old_layout)
        self._container.setLayout(layout)

    def _on_slider_change(self, value):
        self._readout_widget.setValue(self._post_get_hook(value))

    def _on_readout_change(self):
        self._qwidget.setValue(self._pre_set_hook(self._readout_widget.value()))

    def _mgui_get_native_widget(self) -> QtW.QWidget:
        return self._container

    def _mgui_set_min(self, value: float):
        """Set the minimum possible value."""
        super()._mgui_set_min(value)
        self._readout_widget.setMinimum(self._cast(value))

    def _mgui_set_max(self, value: float):
        """Set the maximum possible value."""
        super()._mgui_set_max(value)
        self._readout_widget.setMaximum(self._cast(value))

    def _mgui_set_step(self, value: float):
        """Set the step size."""
        super()._mgui_set_step(value)
        self._readout_widget.setSingleStep(self._cast(value))

    def _mgui_get_adaptive_step(self) -> bool:
        return (
            self._readout_widget.stepType()
            == QtW.QAbstractSpinBox.StepType.AdaptiveDecimalStepType
        )

    def _mgui_set_adaptive_step(self, value: bool):
        self._readout_widget.setStepType(
            QtW.QAbstractSpinBox.StepType.AdaptiveDecimalStepType
            if value
            else QtW.QAbstractSpinBox.StepType.DefaultStepType
        )

    def _mgui_set_readout_visibility(self, value: bool):
        self._readout_widget.show() if value else self._readout_widget.hide()


class FloatSlider(Slider):
    _readout = QtW.QDoubleSpinBox
    _precision = 1e6

    def _update_precision(self, minimum=None, maximum=None, step=None):
        """Called when min/max/step is changed.

        _precision is the factor that converts from integer representation in the slider
        widget, to the actual float representation needed.
        """
        orig = self._precision

        if minimum is not None or maximum is not None:
            _min = minimum or self._mgui_get_min()
            _max = maximum or self._mgui_get_max()

            # make sure val * precision is within int32 overflow limit for Qt
            val = max([abs(_min), abs(_max)])
            while abs(self._precision * val) >= 2**32 // 2:
                self._precision *= 0.1
        elif step:
            while step < (1 / self._precision):
                self._precision *= 10

        ratio = self._precision / orig
        if ratio != 1:
            self._mgui_set_value(self._mgui_get_value() * ratio)
            if not step:
                self._mgui_set_max(self._mgui_get_max() * ratio)
                self._mgui_set_min(self._mgui_get_min() * ratio)
            # self._mgui_set_step(self._mgui_get_step() * ratio)

    def _cast(self, value: Any) -> Any:
        return float(value)

    def _post_get_hook(self, value):
        return value / self._precision

    def _pre_set_hook(self, value):
        return int(value * self._precision)

    def _mgui_bind_change_callback(self, callback):
        def _converted_value(value):
            callback(self._post_get_hook(value))

        self._qwidget.valueChanged.connect(_converted_value)


class RangeSlider(_Slider):
    _qwidget: superqt.QLabeledRangeSlider

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(superqt.QLabeledRangeSlider, **kwargs)
        if hasattr(self._qwidget, "applyMacStylePatch"):
            # >= magicgui v0.5.2
            self._qwidget.applyMacStylePatch()

    def _mgui_set_readout_visibility(self, value: bool):
        method = "show" if value else "hide"
        try:
            for label in chain(
                self._qwidget._handle_labels,
                [self._qwidget._min_label, self._qwidget._max_label],
            ):
                getattr(label, method)()
        except AttributeError as e:
            warnings.warn(str(e), stacklevel=2)

    def _mgui_set_adaptive_step(self, value: bool):
        pass

    def _mgui_get_adaptive_step(self) -> bool:
        return False

    def _pre_set_hook(self, value):
        return value


class FloatRangeSlider(RangeSlider):
    _qwidget: superqt.QLabeledDoubleRangeSlider

    def __init__(self, **kwargs: Any) -> None:
        _Slider.__init__(self, superqt.QLabeledDoubleRangeSlider, **kwargs)
        if hasattr(self._qwidget, "applyMacStylePatch"):
            # >= magicgui v0.5.2
            self._qwidget.applyMacStylePatch()


class ProgressBar(_Slider):
    _qwidget: QtW.QProgressBar

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QProgressBar, **kwargs)

    def _mgui_get_step(self) -> float:
        """Get the step size."""
        return 1

    def _mgui_set_step(self, value: float):
        """Set the step size."""

    def _mgui_set_adaptive_step(self, value: bool):
        """Set is step is adaptive."""

    def _mgui_get_adaptive_step(self) -> bool:
        return False

    def _mgui_set_readout_visibility(self, value: bool):
        self._qwidget.setTextVisible(value)


class ComboBox(QBaseValueWidget, protocols.CategoricalWidgetProtocol):
    _qwidget: QtW.QComboBox

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QComboBox, "isChecked", "setCurrentIndex", "", **kwargs)
        self._qwidget.currentIndexChanged.connect(self._emit_data)

    def _emit_data(self, index: int):
        self._event_filter.valueChanged.emit(self._qwidget.itemData(index))

    def _mgui_bind_change_callback(self, callback):
        self._event_filter.valueChanged.connect(callback)

    def _mgui_get_count(self) -> int:
        """Return the number of items in the dropdown, omitting any separator items."""
        return sum(
            1
            for i in range(self._qwidget.count())
            if self._qwidget.itemData(i) != Separator
        )

    def _mgui_get_choice(self, choice_name: str) -> Any:
        item_index = self._qwidget.findText(choice_name)
        return None if item_index == -1 else self._qwidget.itemData(item_index)

    def _mgui_get_current_choice(self) -> str:
        return self._qwidget.itemText(self._qwidget.currentIndex())

    def _mgui_get_value(self) -> Any:
        return self._qwidget.itemData(self._qwidget.currentIndex())

    def _mgui_set_value(self, value) -> None:
        # Note: there's a bug in PyQt6, where CombBox.findData(value) will not
        # find the data if value is an Enum. So we do it manually
        wdg = self._qwidget
        idx = next((i for i in range(wdg.count()) if wdg.itemData(i) == value), -1)
        self._qwidget.setCurrentIndex(idx)

    def _mgui_set_choice(self, choice_name: str, data: Any) -> None:
        """Set data for ``choice_name``."""
        if data is Separator:
            item_index = self._qwidget.count()
            self._qwidget.insertSeparator(item_index)  # itemData is None
            self._qwidget.setItemData(item_index, Separator)
        else:
            item_index = self._qwidget.findText(choice_name)
            # if it's not in the list, add a new item
            if item_index == -1:
                self._qwidget.addItem(choice_name, data)
            # otherwise update its data
            else:
                self._qwidget.setItemData(item_index, data)

    def _mgui_set_choices(self, choices: Iterable[tuple[str, Any]]) -> None:
        """Set current items in categorical type ``widget`` to ``choices``."""
        choices_ = list(choices)
        if not choices_:
            self._qwidget.clear()
            return

        with _signals_blocked(self._qwidget):
            choice_names = [x[0] for x in choices_]
            # remove choices that no longer exist
            current = self._qwidget.itemText(self._qwidget.currentIndex())
            for i in reversed(range(self._qwidget.count())):
                if self._qwidget.itemText(i) not in choice_names:
                    self._qwidget.removeItem(i)
            # update choices and insert separators
            for name, data in choices_:
                self._mgui_set_choice(name, data)
            # if the currently selected item is not in the new set,
            # remove it and select the first item in the list
            current2 = self._qwidget.itemText(self._qwidget.currentIndex())
            if current not in choice_names:
                # previous value was not in the new choices so set first element
                first = choice_names[0]
                self._qwidget.setCurrentIndex(self._qwidget.findText(first))
            elif current2 != current:
                # element is present but order is different
                self._qwidget.setCurrentIndex(self._qwidget.findText(current))
        if current not in choice_names:
            self._emit_data(self._qwidget.currentIndex())

    def _mgui_del_choice(self, choice_name: str) -> None:
        """Delete choice_name."""
        item_index = self._qwidget.findText(choice_name)
        if item_index >= 0:
            self._qwidget.removeItem(item_index)

    def _mgui_get_choices(self) -> tuple[tuple[str, Any], ...]:
        """Get available choices."""
        return tuple(
            (self._qwidget.itemText(i), self._qwidget.itemData(i))
            for i in range(self._qwidget.count())
            if self._qwidget.itemData(i) is not Separator
        )


class Select(QBaseValueWidget, protocols.CategoricalWidgetProtocol):
    _qwidget: QtW.QListWidget

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QListWidget, "isChecked", "setCurrentIndex", "", **kwargs)
        self._qwidget.itemSelectionChanged.connect(self._emit_data)
        self._qwidget.setSelectionMode(
            QtW.QAbstractItemView.SelectionMode.ExtendedSelection
        )

    def _emit_data(self):
        data = self._qwidget.selectedItems()
        self._event_filter.valueChanged.emit(
            [d.data(Qt.ItemDataRole.UserRole) for d in data]
        )

    def _mgui_bind_change_callback(self, callback):
        self._event_filter.valueChanged.connect(callback)

    def _mgui_get_count(self) -> int:
        """Return the number of items in the dropdown."""
        return self._qwidget.count()

    def _mgui_get_choice(self, choice_name: str) -> list[Any]:
        items = self._qwidget.findItems(choice_name, Qt.MatchExactly)
        return [i.data(Qt.ItemDataRole.UserRole) for i in items]

    def _mgui_get_current_choice(self) -> list[str]:  # type: ignore[override]
        return [i.text() for i in self._qwidget.selectedItems()]

    def _mgui_get_value(self) -> Any:
        return [i.data(Qt.ItemDataRole.UserRole) for i in self._qwidget.selectedItems()]

    def _mgui_set_value(self, value) -> None:
        if not isinstance(value, (list, tuple)):
            value = [value]
        selected_prev = self._qwidget.selectedItems()
        with _signals_blocked(self._qwidget):
            for i in range(self._qwidget.count()):
                item = self._qwidget.item(i)
                item.setSelected(item.data(Qt.ItemDataRole.UserRole) in value)
        selected_post = self._qwidget.selectedItems()
        if selected_prev != selected_post:
            self._emit_data()

    def _mgui_set_choice(self, choice_name: str, data: Any) -> None:
        """Set data for ``choice_name``."""
        items = self._qwidget.findItems(choice_name, Qt.MatchFlag.MatchExactly)
        # if it's not in the list, add a new item
        if not items:
            item = QtW.QListWidgetItem(choice_name)
            item.setData(Qt.ItemDataRole.UserRole, data)
            self._qwidget.addItem(item)
        # otherwise update its data
        else:
            for item in items:
                item.setData(Qt.ItemDataRole.UserRole, data)

    def _mgui_set_choices(self, choices: Iterable[tuple[str, Any]]) -> None:
        """Set current items in categorical type ``widget`` to ``choices``."""
        choices_ = list(choices)
        if not choices_:
            self._qwidget.clear()
            return

        with _signals_blocked(self._qwidget):
            choice_names = [x[0] for x in choices_]
            selected_prev = self._qwidget.selectedItems()
            # remove choices that no longer exist
            for i in reversed(range(self._qwidget.count())):
                if self._qwidget.item(i).text() not in choice_names:
                    self._qwidget.takeItem(i)
            # update choices
            for name, data in choices_:
                self._mgui_set_choice(name, data)
            selected_post = self._qwidget.selectedItems()
        if selected_prev != selected_post:
            self._emit_data()

    def _mgui_del_choice(self, choice_name: str) -> None:
        """Delete choice_name."""
        for i in reversed(range(self._qwidget.count())):
            if self._qwidget.item(i).text() == choice_name:
                self._qwidget.takeItem(i)

    def _mgui_get_choices(self) -> tuple[tuple[str, Any], ...]:
        """Get available choices."""
        return tuple(
            (
                self._qwidget.item(i).text(),
                self._qwidget.item(i).data(Qt.ItemDataRole.UserRole),
            )
            for i in range(self._qwidget.count())
        )


class RadioButtons(
    QBaseValueWidget,
    protocols.CategoricalWidgetProtocol,
    protocols.SupportsOrientation,
):
    _qwidget: QtW.QGroupBox

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QGroupBox, "", "", "", **kwargs)
        self._btn_group = QtW.QButtonGroup(self._qwidget)
        self._mgui_set_orientation("vertical")
        self._btn_group.buttonToggled.connect(self._emit_data)

    def _emit_data(self, btn, checked):
        if checked:
            self._event_filter.valueChanged.emit(self._mgui_get_value())

    def _mgui_bind_change_callback(self, callback):
        self._event_filter.valueChanged.connect(callback)

    def _mgui_set_orientation(self, value: str) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        new_layout = QtW.QHBoxLayout() if value == "horizontal" else QtW.QVBoxLayout()
        for btn in self._btn_group.buttons():
            new_layout.addWidget(btn)
        old_layout = self._qwidget.layout()
        if old_layout is not None:
            QtW.QWidget().setLayout(old_layout)
        self._qwidget.setLayout(new_layout)
        self._qwidget.layout().setContentsMargins(4, 4, 4, 4)

    def _mgui_get_orientation(self) -> str:
        """Get orientation, return either 'horizontal' or 'vertical'."""
        if isinstance(self._qwidget.layout(), QtW.QVBoxLayout):
            return "vertical"
        else:
            return "horizontal"

    def _mgui_get_current_choice(self) -> str:
        btn = self._btn_group.checkedButton()
        return btn.text() if btn else None

    def _mgui_get_value(self) -> Any:
        btn = self._btn_group.checkedButton()
        return btn._data if btn else None

    def _mgui_set_value(self, value) -> None:
        for btn in self._btn_group.buttons():
            if btn._data == value:
                btn.setChecked(True)
                break  # exclusive

    def _mgui_get_count(self) -> int:
        """Return the number of items in the dropdown."""
        return len(self._btn_group.buttons())

    def _mgui_get_choice(self, choice_name: str) -> Any:
        for btn in self._btn_group.buttons():
            if btn.text() == choice_name:
                return btn._data
        return None

    def _add_button(self, label: str, data: Any | None = None):
        btn = QtW.QRadioButton(label, self._qwidget)
        btn._data = data
        self._btn_group.addButton(btn)
        self._qwidget.layout().addWidget(btn)

    def _remove_button(self, btn):
        self._btn_group.removeButton(btn)
        btn.setParent(None)
        btn.deleteLater()

    def _mgui_set_choice(self, choice_name: str, data: Any) -> None:
        """Set data for ``choice_name``."""
        for btn in self._btn_group.buttons():
            if btn.text() == choice_name:
                # otherwise update its data
                btn._data = data
        else:
            # if it's not in the list, add a new item
            self._add_button(choice_name, data)

    def _mgui_del_choice(self, choice_name: str) -> None:
        """Delete choice_name."""
        for btn in self._btn_group.buttons():
            if btn.text() == choice_name:
                self._remove_button(btn)
                break

    def _mgui_get_choices(self) -> tuple[tuple[str, Any], ...]:
        """Get available choices."""
        return tuple((str(btn.text()), btn._data) for btn in self._btn_group.buttons())

    def _mgui_set_choices(self, choices: Iterable[tuple[str, Any]]) -> None:
        """Set current items in categorical type ``widget`` to ``choices``."""
        current = self._mgui_get_value()
        for btn in self._btn_group.buttons():
            self._remove_button(btn)

        for c in choices:
            self._add_button(*c)
        self._mgui_set_value(current)


class DateTimeEdit(QBaseValueWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            QtW.QDateTimeEdit, "", "setDateTime", "dateTimeChanged", **kwargs
        )

    def _mgui_get_value(self):
        try:
            return self._qwidget.dateTime().toPython()
        except (TypeError, AttributeError):
            return self._qwidget.dateTime().toPyDateTime()


class DateEdit(QBaseValueWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QDateEdit, "", "setDate", "dateChanged", **kwargs)

    def _mgui_get_value(self):
        try:
            return self._qwidget.date().toPython()
        except (TypeError, AttributeError):
            return self._qwidget.date().toPyDate()


class TimeEdit(QBaseValueWidget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QTimeEdit, "", "setTime", "timeChanged", **kwargs)

    def _mgui_get_value(self):
        try:
            return self._qwidget.time().toPython()
        except (TypeError, AttributeError):
            return self._qwidget.time().toPyTime()


class ToolBar(QBaseWidget):
    _qwidget: QtW.QToolBar

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(QtW.QToolBar, **kwargs)
        self._qwidget.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self._event_filter.paletteChanged.connect(self._on_palette_change)

    def _on_palette_change(self):
        for action in self._qwidget.actions():
            if icon := action.data():
                if qicon := _get_qicon(icon, None, palette=self._qwidget.palette()):
                    action.setIcon(qicon)

    def _mgui_add_button(self, text: str, icon: str, callback: Callable) -> None:
        """Add an action to the toolbar."""
        act = self._qwidget.addAction(text, callback)
        if qicon := _get_qicon(icon, None, palette=self._qwidget.palette()):
            act.setIcon(qicon)
            act.setData(icon)

    def _mgui_add_separator(self) -> None:
        """Add a separator line to the toolbar."""
        self._qwidget.addSeparator()

    def _mgui_add_spacer(self) -> None:
        """Add a spacer to the toolbar."""
        empty = QtW.QWidget()
        empty.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Preferred
        )
        self._qwidget.addWidget(empty)

    def _mgui_add_widget(self, widget: Widget) -> None:
        """Add a widget to the toolbar."""
        self._qwidget.addWidget(widget.native)

    def _mgui_get_icon_size(self) -> tuple[int, int] | None:
        """Return the icon size of the toolbar."""
        sz = self._qwidget.iconSize()
        return None if sz.isNull() else (sz.width(), sz.height())

    def _mgui_set_icon_size(self, size: int | tuple[int, int] | None) -> None:
        """Set the icon size of the toolbar."""
        if isinstance(size, int):
            _size = QSize(size, size)
        elif isinstance(size, tuple):
            _size = QSize(size[0], size[1])
        else:
            _size = QSize()
        self._qwidget.setIconSize(_size)

    def _mgui_clear(self) -> None:
        """Clear the toolbar."""
        self._qwidget.clear()


class Dialog(QBaseWidget, protocols.ContainerProtocol):
    def __init__(
        self, layout="vertical", scrollable: bool = False, **kwargs: Any
    ) -> None:
        QBaseWidget.__init__(self, QtW.QDialog, **kwargs)
        if layout == "horizontal":
            self._layout: QtW.QBoxLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()
        self._qwidget.setLayout(self._layout)
        self._layout.setSizeConstraint(QtW.QLayout.SizeConstraint.SetMinAndMaxSize)

        button_box = QtW.QDialogButtonBox(
            QtW.QDialogButtonBox.StandardButton.Ok
            | QtW.QDialogButtonBox.StandardButton.Cancel,
            Qt.Orientation.Horizontal,
            self._qwidget,
        )
        button_box.accepted.connect(self._qwidget.accept)
        button_box.rejected.connect(self._qwidget.reject)
        self._layout.addWidget(button_box)

    def _mgui_insert_widget(self, position: int, widget: Widget):
        self._layout.insertWidget(position, widget.native)

    def _mgui_remove_widget(self, widget: Widget):
        self._layout.removeWidget(widget.native)
        widget.native.setParent(None)

    def _mgui_get_margins(self) -> tuple[int, int, int, int]:
        m = self._layout.contentsMargins()
        return m.left(), m.top(), m.right(), m.bottom()

    def _mgui_set_margins(self, margins: tuple[int, int, int, int]) -> None:
        self._layout.setContentsMargins(*margins)

    def _mgui_set_orientation(self, value) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        raise NotImplementedError("Setting orientation not supported for dialogs.")

    def _mgui_get_orientation(self) -> str:
        """Set orientation, return either 'horizontal' or 'vertical'."""
        return "horizontal" if isinstance(self, QtW.QHBoxLayout) else "vertical"

    def _mgui_exec(self) -> Any:
        return self._qwidget.exec_()


QFILE_DIALOG_MODES = {
    FileDialogMode.EXISTING_FILE: QtW.QFileDialog.getOpenFileName,
    FileDialogMode.EXISTING_FILES: QtW.QFileDialog.getOpenFileNames,
    FileDialogMode.OPTIONAL_FILE: QtW.QFileDialog.getSaveFileName,
    FileDialogMode.EXISTING_DIRECTORY: QtW.QFileDialog.getExistingDirectory,
}


def show_file_dialog(
    mode: str | FileDialogMode = FileDialogMode.EXISTING_FILE,
    caption: str | None = None,
    start_path: str | None = None,
    filter: str | None = None,
    parent=None,
) -> str | None:
    show_dialog = QFILE_DIALOG_MODES[FileDialogMode(mode)]
    if FileDialogMode(mode) is FileDialogMode.EXISTING_DIRECTORY:
        result = show_dialog(parent, caption, start_path)
    else:
        result, _ = show_dialog(parent, caption, start_path, filter)
    return result or None


def _might_be_rich_text(text):
    return bool(re.search("<[^\n]+>", text))


def get_text_width(text: str) -> int:
    """Return the width required to render ``text`` (including rich text elements)."""
    if _might_be_rich_text(text):
        doc = QTextDocument()
        doc.setHtml(text)
        return math.ceil(doc.size().width())
    else:
        fm = QFontMetrics(QFont("", 0))
        return fm.boundingRect(text).width() + 5


def _maybefloat(item):
    if not item:
        return None
    num = item.text() if hasattr(item, "text") else item
    try:
        return int(num) if num.isdigit() else float(num)
    except ValueError:
        return num


class _QTableExtended(QtW.QTableWidget):
    _read_only: bool = False

    def __init__(self, *args, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.setItemDelegate(_ItemDelegate(parent=self))

    def _copy_to_clipboard(self):
        selranges = self.selectedRanges()
        if not selranges:
            return
        if len(selranges) > 1:
            import warnings

            warnings.warn(
                "Multiple table selections detected: "
                "only the first (upper left) selection will be copied",
                stacklevel=2,
            )

        # copy first selection range
        sel = selranges[0]
        lines = []
        for r in range(sel.topRow(), sel.bottomRow() + 1):
            cells = []
            for c in range(sel.leftColumn(), sel.rightColumn() + 1):
                item = self.item(r, c)
                cells.append(item.text()) if hasattr(item, "text") else ""
            lines.append("\t".join(cells))

        if lines:
            QtW.QApplication.clipboard().setText("\n".join(lines))

    def _paste_from_clipboard(self):
        if self._read_only:
            return

        sel_idx = self.selectedIndexes()
        if not sel_idx:
            return
        text = QtW.QApplication.clipboard().text()
        if not text:
            return

        # paste in the text
        row0, col0 = sel_idx[0].row(), sel_idx[0].column()
        data = [line.split("\t") for line in text.splitlines()]
        if (row0 + len(data)) > self.rowCount():
            self.setRowCount(row0 + len(data))
        if data and (col0 + len(data[0])) > self.columnCount():
            self.setColumnCount(col0 + len(data[0]))
        for r, line in enumerate(data):
            for c, cell in enumerate(line):
                try:
                    self.item(row0 + r, col0 + c).setText(str(cell))
                except AttributeError:
                    self.setItem(row0 + r, col0 + c, QtW.QTableWidgetItem(str(cell)))

        # select what was just pasted
        selrange = QtW.QTableWidgetSelectionRange(row0, col0, row0 + r, col0 + c)
        self.clearSelection()
        self.setRangeSelected(selrange, True)

    def _delete_selection(self):
        if self._read_only:
            return

        for item in self.selectedItems():
            with suppress(AttributeError):
                item.setText("")

    def keyPressEvent(self, e: QKeyEvent):
        if e.modifiers() & Qt.ControlModifier and e.key() == Qt.Key_C:
            return self._copy_to_clipboard()
        if e.modifiers() & Qt.ControlModifier and e.key() == Qt.Key_V:
            return self._paste_from_clipboard()
        if e.modifiers() & Qt.ControlModifier and e.key() == Qt.Key_X:
            self._copy_to_clipboard()
            return self._delete_selection()
        if e.key() in (Qt.Key_Delete, Qt.Key_Backspace):
            return self._delete_selection()
        return super().keyPressEvent(e)


_READ_ONLY = QtW.QTableWidget.EditTrigger.NoEditTriggers
_EDITABLE = (
    QtW.QTableWidget.EditTrigger.EditKeyPressed
    | QtW.QTableWidget.EditTrigger.DoubleClicked
)
_DATA_ROLE: int = 255


class Table(QBaseWidget, protocols.TableWidgetProtocol):
    _qwidget: _QTableExtended

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(_QTableExtended, **kwargs)
        header = self._qwidget.horizontalHeader()
        # avoid strange AttributeError on CI
        if hasattr(header, "setSectionResizeMode"):
            header.setSectionResizeMode(QtW.QHeaderView.ResizeMode.Stretch)
        # self._qwidget.horizontalHeader().setSectionsMovable(True)  # tricky!!
        header.setSectionResizeMode(QtW.QHeaderView.ResizeMode.Interactive)
        self._qwidget.itemChanged.connect(self._update_item_data_with_text)

    def _mgui_set_read_only(self, value: bool) -> None:
        value = bool(value)
        self._qwidget._read_only = value
        if value:
            self._qwidget.setEditTriggers(_READ_ONLY)
        else:
            self._qwidget.setEditTriggers(_EDITABLE)

    def _mgui_get_read_only(self) -> bool:
        return self._qwidget._read_only

    def _update_item_data_with_text(self, item: QtW.QTableWidgetItem):
        with _signals_blocked(self._qwidget):
            item.setData(_DATA_ROLE, _maybefloat(item.text()))

    def _mgui_set_row_count(self, nrows: int) -> None:
        """Set the number of rows in the table. (Create/delete as needed)."""
        self._qwidget.setRowCount(nrows)

    def _mgui_set_column_count(self, ncols: int) -> None:
        """Set the number of columns in the table. (Create/delete as needed)."""
        self._qwidget.setColumnCount(ncols)

    def _mgui_get_column_count(self) -> int:
        return self._qwidget.columnCount()

    def _mgui_get_row_count(self) -> int:
        return self._qwidget.rowCount()

    def _mgui_remove_row(self, row: int) -> None:
        self._qwidget.removeRow(row)

    def _mgui_remove_column(self, column: int) -> None:
        self._qwidget.removeColumn(column)

    def _mgui_get_cell(self, row: int, col: int) -> Any:
        """Get current value of the widget."""
        item = self._qwidget.item(row, col)
        if item:
            return item.data(_DATA_ROLE)
        widget = self._qwidget.cellWidget(row, col)
        if widget:
            return getattr(widget, "_magic_widget", widget)

    def _mgui_set_cell(self, row: int, col: int, value: Any) -> None:
        """Set current value of the widget."""
        if value is None:
            self._qwidget.setItem(row, col, None)
            self._qwidget.removeCellWidget(row, col)
            return
        if isinstance(value, Widget):
            self._qwidget.setCellWidget(row, col, value.native)
            return
        item = QtW.QTableWidgetItem(str(value))
        item.setData(_DATA_ROLE, value)
        self._qwidget.setItem(row, col, item)

    def _mgui_get_row_headers(self) -> tuple:
        """Get current row headers of the widget."""
        # ... 'PySide2.QtWidgets.QTableWidgetItem' object has no attribute 'visualIndex'
        # on some linux... so removing the visualIndex() for row headers... this means
        # we can't enable row drag/drop without fixing this
        indices = range(self._qwidget.rowCount())
        items = (self._qwidget.verticalHeaderItem(i) for i in indices)
        headers = (item.text() for item in items if item)
        return tuple(_maybefloat(x) for x in headers)

    def _mgui_set_row_headers(self, headers: Sequence) -> None:
        """Set current row headers of the widget."""
        self._qwidget.setVerticalHeaderLabels(tuple(map(str, headers)))

    def _mgui_get_column_headers(self) -> tuple:
        """Get current column headers of the widget."""
        horiz_header = self._qwidget.horizontalHeader()
        ncols = self._qwidget.columnCount()
        # visual index allows for column drag/drop, though it's currently not enabled.
        indices = (horiz_header.visualIndex(i) for i in range(ncols))
        items = (self._qwidget.horizontalHeaderItem(i) for i in indices)
        headers = (item.text() for item in items if item)
        return tuple(_maybefloat(x) for x in headers)

    def _mgui_set_column_headers(self, headers: Sequence) -> None:
        """Set current column headers of the widget."""
        self._qwidget.setHorizontalHeaderLabels(tuple(map(str, headers)))

    def _mgui_bind_row_headers_change_callback(self, callback) -> None:
        """Bind callback to row headers change event."""
        raise NotImplementedError()

    def _mgui_bind_column_headers_change_callback(self, callback) -> None:
        """Bind callback to column headers change event."""
        raise NotImplementedError()

    def _mgui_bind_change_callback(self, callback):
        """Bind callback to event of changing any cell."""
        self._qwidget.itemChanged.connect(partial(self._item_callback, callback))

    def _item_callback(self, callback, item: QtW.QTableWidgetItem):
        col_head = item.tableWidget().horizontalHeaderItem(item.column())
        col_head = col_head.text() if col_head is not None else ""
        row_head = item.tableWidget().verticalHeaderItem(item.row())
        row_head = row_head.text() if row_head is not None else ""
        data = {
            "data": item.data(_DATA_ROLE),
            "row": item.row(),
            "column": item.column(),
            "column_header": col_head,
            "row_header": row_head,
        }
        callback(data)

    # These are only here to implement the ValueWidget interface... but in this one
    # case, all of the get/set value logic happens in magicgui.widgets.Table
    # calling Table._mgui_set_cell and Table._mgui_get_cell instead
    def _mgui_get_value(self):
        pass

    def _mgui_set_value(self):
        pass


class _ItemDelegate(QtW.QStyledItemDelegate):
    """Displays table widget items with properly formatted numbers."""

    def displayText(self, value, locale):
        return super().displayText(_format_number(value, 4), locale)


def _format_number(text: str, ndigits: int = 4) -> str:
    """Convert string to int or float if possible."""
    try:
        value: int | float | None = int(text)
    except ValueError:
        try:
            value = float(text)
        except ValueError:
            value = None

    if isinstance(value, (int, float)):
        if 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
            text = str(value) if isinstance(value, int) else f"{value:.{ndigits}f}"
        else:
            text = f"{value:.{ndigits-1}e}"

    return text



"""Tests for the tqdm wrapper."""

from time import sleep

import pytest

from magicgui import magicgui

tqdm = pytest.importorskip("tqdm", reason="need tqdm installed to test tqdm")
from magicgui.tqdm import tqdm, trange  # noqa


def test_tqdm_outside_of_functiongui():
    """Test that we can make a tqdm wrapper with ProgressBar outside of @magicgui."""
    with trange(10) as pbar:
        assert pbar.n == 0
        assert pbar.total == 10
        assert not pbar._mgui

    assert tuple(trange(5)) == tuple(range(5))


def test_indeterminate_tqdm_outside_of_functiongui():
    """Test that we can have tqdm with an indeterminate total range."""
    with tqdm(total=None) as pbar:
        assert pbar.total is None
        assert pbar.n == 0
        pbar.update()
        assert pbar.n == 1


def test_disabled_tqdm():
    """Test that a disabled tqdm does not have a progressbar or magicgui."""

    @magicgui
    def f():
        with trange(10, disable=True) as pbar:
            assert not hasattr(pbar, "progressbar")
            assert not pbar._mgui


def test_no_leave_tqdm():
    """Test that leave=False hides the progressbar when done (tqdm API wording)"""

    @magicgui
    def f():
        with trange(10, leave=True) as pbar1:
            pass
        assert pbar1.progressbar.visible is True

    f.show()
    f()

    @magicgui
    def f2():
        with trange(10, leave=False) as pbar2:
            pass
        assert pbar2.progressbar.visible is False

    f2.show()
    f2()


def test_unbound_tqdm():
    """Test that tqdm without defined total sill have a range of (0, 0)."""

    @magicgui
    def f():
        with tqdm() as pbar:
            assert pbar.total is None
            # undefined range will render as a "busy" indicator
            assert pbar.progressbar.range == (0, 0)

    f.show()
    f()


def test_tqdm_std_err(capsys):
    """Test that tqdm inside an invisible magicgui falls back to console behavior."""

    # outside of a magicgui it falls back to tqdm_std behavior
    with tqdm(range(10)) as t_obj:
        iter(t_obj)
    captured = capsys.readouterr()

    assert "0%|" in str(captured.err)
    assert "| 0/10" in str(captured.err)
    assert not str(captured.out)


def test_tqdm_in_visible_mgui_std_err(capsys):
    """Test that tqdm inside an mgui outputs to console only when mgui not visible."""

    # inside of a of a magicgui it falls back to tqdm_std behavior
    @magicgui
    def f():
        with tqdm(range(10)) as t_obj:
            iter(t_obj)

    f()
    captured = capsys.readouterr()
    assert "0%|" in str(captured.err)
    assert "| 0/10" in str(captured.err)
    assert not str(captured.out)

    # showing the widget disables the output to console behavior
    f.show()
    f()
    captured = capsys.readouterr()
    assert not str(captured.err)
    assert not str(captured.out)


# Test various ways that tqdm might need to traverse the frame stack:


def test_tqdm_inside_of_magicgui():
    """Test that tqdm can find the magicgui within which it is called."""

    @magicgui
    def long_func(steps=2):
        for _i in tqdm([0, 1]):
            sleep(0.02)

    # before calling the function, we won't have any progress bars
    long_func.show()
    assert not long_func._tqdm_pbars
    long_func()
    # after calling the it, we should now have a progress bars
    assert len(long_func._tqdm_pbars) == 1
    # but calling it again won't add more
    long_func()
    assert len(long_func._tqdm_pbars) == 1


def test_trange_inside_of_magicgui():
    """Test that trange can find the magicgui within which it is called."""

    @magicgui
    def long_func(steps=2):
        for _i in trange(4):
            pass

    long_func.show()
    assert not long_func._tqdm_pbars
    long_func()
    assert len(long_func._tqdm_pbars) == 1


def test_indeterminate_tqdm_inside_magicgui():
    """Test that we can have tqdm with an indeterminate total range."""

    @magicgui
    def long_func():
        """Long running computation with indeterminate range."""
        with tqdm(total=None):
            pass

    long_func.show()
    assert not long_func._tqdm_pbars
    long_func()
    assert len(long_func._tqdm_pbars) == 1


def _indirectly_decorated(steps=2):
    for _i in trange(4):
        pass


def test_trange_inside_of_indirectly_decorated_magicgui():
    """Test that trange can find the magicgui within which it is called."""
    indirectly_decorated = magicgui(_indirectly_decorated)
    indirectly_decorated.show()
    assert not indirectly_decorated._tqdm_pbars
    indirectly_decorated()
    assert len(indirectly_decorated._tqdm_pbars) == 1


def test_tqdm_nested():
    """Test that tqdm can find the magicgui within which it is called."""

    @magicgui
    def long_func():
        for _i in trange(4):
            for _x in trange(4):
                pass

    long_func.show()
    # before calling the function, we won't have any progress bars
    assert not long_func._tqdm_pbars
    long_func()
    # after calling the it, we should now have a progress bars
    assert len(long_func._tqdm_pbars) == 2

    # the depth is all that matters... not the total number of tqdms
    @magicgui
    def long_func2():
        for _i in trange(4):
            for _x in trange(4):
                pass

        for _x in trange(4):
            pass

    long_func2.show()
    # before calling the function, we won't have any progress bars
    assert not long_func2._tqdm_pbars
    long_func2()
    # after calling the it, we should now have a progress bars
    assert len(long_func2._tqdm_pbars) == 2



"""# Simple progress bar

A simple progress bar demo with magicgui.
"""

from time import sleep

from magicgui import magicgui
from magicgui.tqdm import trange

# if magicui.tqdm.tqdm or trange are used outside of a @magicgui function, (such as in
# interactive use in IPython), then they fall back to the standard terminal output


# If use inside of a magicgui-decorated function
# a progress bar widget will be added to the magicgui container
@magicgui(call_button=True, layout="horizontal")
def long_running(steps=10, delay=0.1):
    """Long running computation with range iterator."""
    # trange(steps) is a shortcut for `tqdm(range(steps))`
    for _i in trange(steps):
        sleep(delay)


long_running.show(run=True)



"""# napari forward reference demo

Example of using a ForwardRef to avoid importing a module that provides a widget.

In this example, one might want to create a widget that takes as an argument a napari
Image layer, and returns an Image.  In order to avoid needing to import napari (and
therefore depending directly on napari), it's possible to annotate those parameters with
a *string* representation of the type (rather than the type itself).  This is called a
"forward reference": https://www.python.org/dev/peps/pep-0484/#forward-references
"""

# Note: if you'd like to avoid circular imports, or just want to avoid having your
# linter yell at you for an undefined type annotation, you can place the import
# inside of an `if typing.TYPE_CHECKING` conditional.  This is not evaluated at runtime,
# only when something like mypy is doing type checking.
from typing import TYPE_CHECKING

from magicgui import magicgui

if TYPE_CHECKING:
    import napari


@magicgui(call_button="execute", background={"max": 200})
def subtract_background(
    data: "napari.types.ImageData", background: int = 50
) -> "napari.types.ImageData":
    """Subtract a constant from the data."""
    if data:
        return data - background


subtract_background.show(run=True)
# now, this example isn't all that interesting on its own (since there will be no Image
# layer in the dropdown) ... but in another package, where you DO import napari,
# you could add this widget to a napari viewer with
# viewer.window.add_dock_widget(subtract_background)



"""magicgui is a utility for generating a GUI from a python function."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("magicgui")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"

__author__ = """Talley Lambert"""
__email__ = "talley.lambert@gmail.com"


from .application import event_loop, use_app
from .type_map import register_type, type_registered
from .type_map._magicgui import magic_factory, magicgui

__all__ = [
    "event_loop",
    "magic_factory",
    "magicgui",
    "register_type",
    "type_registered",
    "use_app",
]



import sys

from qtpy.QtCore import QCoreApplication, Qt, QTimer
from qtpy.QtWidgets import QApplication

try:
    from qtpy import QT6
except ImportError:
    QT6 = False

from magicgui.application import APPLICATION_NAME
from magicgui.widgets.protocols import BaseApplicationBackend


class ApplicationBackend(BaseApplicationBackend):
    _app: QCoreApplication

    def _mgui_get_backend_name(self):
        return "qt"

    def _mgui_process_events(self):
        app = self._mgui_get_native_app()
        app.processEvents()

    def _mgui_run(self):
        app = self._mgui_get_native_app()
        # only start the event loop if magicgui created it
        if app.applicationName() == APPLICATION_NAME:
            return app.exec_()

    def _mgui_quit(self):
        return self._mgui_get_native_app().quit()

    def _mgui_get_native_app(self):
        # Get native app
        self._app = QApplication.instance()
        if not self._app:
            if not QT6:
                high_dpi = Qt.ApplicationAttribute.AA_EnableHighDpiScaling
                QApplication.setAttribute(high_dpi)
            self._app = QApplication(sys.argv)
            self._app.setApplicationName(APPLICATION_NAME)
        return self._app

    def _mgui_start_timer(self, interval=0, on_timeout=None, single=False):
        self._timer = QTimer()
        if on_timeout:
            self._timer.timeout.connect(on_timeout)
        self._timer.setSingleShot(single)
        self._timer.setInterval(interval)
        self._timer.start()

    def _mgui_stop_timer(self):
        if getattr(self, "_timer", None):
            self._timer.stop()



import sys
import typing
from collections.abc import Mapping, Sequence
from concurrent.futures import Future

import pytest

from magicgui._util import safe_issubclass


class TestSafeIsSubclass:
    def test_basic(self):
        assert safe_issubclass(int, int)
        assert safe_issubclass(int, object)

    def test_generic_base(self):
        assert safe_issubclass(typing.List[int], list)
        assert safe_issubclass(typing.List[int], typing.List)

    def test_multiple_generic_base(self):
        assert safe_issubclass(typing.List[int], (typing.List, typing.Dict))

    def test_no_exception(self):
        assert not safe_issubclass(int, 1)

    def test_typing_inheritance(self):
        assert safe_issubclass(typing.List, list)
        assert safe_issubclass(list, typing.List)
        assert safe_issubclass(typing.Tuple, tuple)
        assert safe_issubclass(tuple, typing.Tuple)
        assert safe_issubclass(typing.Dict, dict)
        assert safe_issubclass(dict, typing.Dict)

    def test_inheritance_generic_list(self):
        assert safe_issubclass(list, typing.Sequence)
        assert safe_issubclass(typing.List, typing.Sequence)
        assert safe_issubclass(typing.List[int], typing.Sequence[int])
        assert safe_issubclass(typing.List[int], typing.Sequence)
        assert safe_issubclass(typing.List[int], Sequence)

    def test_no_inheritance_generic_super(self):
        assert not safe_issubclass(list, typing.List[int])

    def test_inheritance_generic_mapping(self):
        assert safe_issubclass(dict, typing.Mapping)
        assert safe_issubclass(typing.Dict, typing.Mapping)
        assert safe_issubclass(typing.Dict[int, str], typing.Mapping[int, str])
        assert safe_issubclass(typing.Dict[int, str], typing.Mapping)
        assert safe_issubclass(typing.Dict[int, str], Mapping)

    @pytest.mark.skipif(sys.version_info < (3, 9), reason="PEP-585 is supported in 3.9+")
    def test_typing_builtins_list(self):
        assert safe_issubclass(list[int], list)
        assert safe_issubclass(list[int], Sequence)
        assert safe_issubclass(list[int], typing.Sequence)
        assert safe_issubclass(list[int], typing.Sequence[int])
        assert safe_issubclass(list[int], typing.List[int])
        assert safe_issubclass(typing.List[int], list)
        assert safe_issubclass(typing.List[int], list[int])

    @pytest.mark.skipif(sys.version_info < (3, 9), reason="PEP-585 is supported in 3.9+")
    def test_typing_builtins_dict(self):
        assert safe_issubclass(dict[int, str], dict)
        assert safe_issubclass(dict[int, str], Mapping)
        assert safe_issubclass(dict[int, str], typing.Mapping)
        assert safe_issubclass(dict[int, str], typing.Mapping[int, str])
        assert safe_issubclass(dict[int, str], typing.Dict[int, str])
        assert safe_issubclass(typing.Dict[int, str], dict)
        assert safe_issubclass(typing.Dict[int, str], dict[int, str])

    def test_tuple_check(self):
        assert safe_issubclass(typing.Tuple[int, str], tuple)
        assert safe_issubclass(typing.Tuple[int], typing.Sequence[int])
        assert safe_issubclass(typing.Tuple[int, int], typing.Sequence[int])
        assert safe_issubclass(typing.Tuple[int, ...], typing.Sequence[int])
        assert safe_issubclass(typing.Tuple[int, ...], typing.Iterable[int])
        assert not safe_issubclass(typing.Tuple[int, ...], typing.Dict[int, typing.Any])
        assert safe_issubclass(typing.Tuple[int, ...], typing.Tuple[int, ...])
        assert safe_issubclass(typing.Tuple[int, int], typing.Tuple[int, ...])
        assert not safe_issubclass(typing.Tuple[int, int], typing.Tuple[int, str])
        assert not safe_issubclass(typing.Tuple[int, int], typing.Tuple[int, int, int])

    @pytest.mark.skipif(sys.version_info < (3, 9), reason="Future[] is supported in 3.9+")
    def test_subclass_future(self):
        assert safe_issubclass(Future[typing.List[int]], Future[list[int]])
        assert safe_issubclass(Future[typing.List[int]], Future[list])
        assert safe_issubclass(Future[list[int]], Future[typing.List[int]])
        assert not safe_issubclass(Future[list[int]], Future[typing.List[str]])



"""Widgets backed by a backend implementation, ready to be instantiated by the user.

All of these widgets should provide the `widget_type` argument to their
super().__init__ calls.
"""

from __future__ import annotations

import datetime
import inspect
import math
import os
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ForwardRef,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from weakref import ref

from typing_extensions import Annotated, get_args, get_origin

from magicgui._type_resolution import resolve_single_type
from magicgui._util import merge_super_sigs, safe_issubclass
from magicgui.application import use_app
from magicgui.types import ChoicesType, FileDialogMode, PathLike, Undefined, _Undefined
from magicgui.widgets.bases import (
    ButtonWidget,
    CategoricalWidget,
    ContainerWidget,
    DialogWidget,
    MainWindowWidget,
    MultiValuedSliderWidget,
    RangedWidget,
    SliderWidget,
    ToolBarWidget,
    TransformedRangedWidget,
    ValueWidget,
    Widget,
    create_widget,
)
from magicgui.widgets.bases._mixins import _OrientationMixin, _ReadOnlyMixin

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from magicgui.widgets import protocols
    from magicgui.widgets.bases._container_widget import ContainerKwargs
    from magicgui.widgets.bases._widget import WidgetKwargs


WidgetVar = TypeVar("WidgetVar", bound=Widget)
WidgetTypeVar = TypeVar("WidgetTypeVar", bound=Type[Widget])
_V = TypeVar("_V")


@overload
def backend_widget(cls: WidgetTypeVar) -> WidgetTypeVar: ...


@overload
def backend_widget(
    cls: Literal[None] = ...,
    *,
    widget_name: str | None = ...,
    transform: Callable[[type], type] | None = ...,
) -> Callable[[WidgetTypeVar], WidgetTypeVar]: ...


def backend_widget(
    cls: WidgetTypeVar | None = None,
    *,
    widget_name: str | None = None,
    transform: Callable[[type], type] | None = None,
) -> WidgetTypeVar | Callable[[WidgetTypeVar], WidgetTypeVar]:
    """Decorate cls to inject the backend widget of the same name.

    The purpose of this decorator is to "inject" the appropriate backend
    `widget_type` argument into the `Widget.__init__` function, according to the
    app currently being used (i.e. returned by `use_app()`).

    Parameters
    ----------
    cls : Type, optional
        The class being decorated, by default None.
    widget_name : str, optional
        The name of the backend widget to wrap. If None, the name of the class being
        decorated is used.  By default None.
    transform : callable, optional
        A optional function that takes a class and returns a class.  May be used
        to transform the characteristics/methods of the class, by default None

    Returns
    -------
    cls : Type
        The final concrete class backed by a backend widget.
    """

    def wrapper(cls: WidgetTypeVar) -> WidgetTypeVar:
        def __init__(self: object, **kwargs: Any) -> None:
            app = use_app()
            assert app.native
            widget = app.get_obj(widget_name or cls.__name__)
            if transform:
                widget = transform(widget)
            kwargs["widget_type"] = widget
            super(cls, self).__init__(**kwargs)  # type: ignore

        cls.__init__ = __init__  # type: ignore
        cls = merge_super_sigs(cls)
        return cls

    return wrapper(cls) if cls else wrapper


@backend_widget
class EmptyWidget(ValueWidget):
    """A base widget with no value.

    This widget is primarily here to serve as a "hidden widget" to which a value or
    callback can be bound.
    """

    _hidden_value = inspect.Parameter.empty

    def get_value(self) -> Any:
        """Return value if one has been manually set... otherwise return Param.empty."""
        return self._hidden_value

    @property
    def value(self) -> Any:
        """Look for a bound value, otherwise fallback to `get_value`."""
        return super().value

    @value.setter
    def value(self, value: Any) -> None:
        self._hidden_value = value

    def __repr__(self) -> str:
        """Return string repr (avoid looking for value)."""
        try:
            name = f"(name={self.name!r})" if self.name else ""
            return f"<{self.widget_type} {name}>"
        except AttributeError:  # pragma: no cover
            return f"<Uninitialized {self.widget_type}>"


@backend_widget
class Label(ValueWidget[str]):
    """A non-editable text display."""


@backend_widget
class LineEdit(ValueWidget[str]):
    """A one-line text editor."""


@backend_widget
class Password(ValueWidget[str]):
    """A one-line text editor that obscures input."""


@backend_widget
class LiteralEvalLineEdit(ValueWidget[str]):
    """A one-line text editor that evaluates strings as python literals."""


@backend_widget
class TextEdit(ValueWidget[str], _ReadOnlyMixin):  # type: ignore
    """A widget to edit and display both plain and rich text."""


@backend_widget
class DateTimeEdit(ValueWidget[datetime.datetime]):
    """A widget for editing dates and times."""


@backend_widget
class DateEdit(ValueWidget[datetime.date]):
    """A widget for editing dates."""


TV = TypeVar("TV", bound=Union[datetime.time, datetime.timedelta])


@backend_widget
class TimeEdit(ValueWidget[TV]):
    """A widget for editing times."""


@backend_widget
class PushButton(ButtonWidget):
    """A clickable command button."""


@backend_widget
class CheckBox(ButtonWidget):
    """A checkbox with a text label."""


@backend_widget
class RadioButton(ButtonWidget):
    """A radio button with a text label."""


@backend_widget
class SpinBox(RangedWidget[int]):
    """A widget to edit an integer with clickable up/down arrows."""


@backend_widget
class FloatSpinBox(RangedWidget[float]):
    """A widget to edit a float with clickable up/down arrows."""


@backend_widget
class ProgressBar(SliderWidget[float]):
    """A progress bar widget."""

    def increment(self, val: float | None = None) -> None:
        """Increase current value by step size, or provided value."""
        self.value = self.get_value() + (val if val is not None else self.step)

    def decrement(self, val: float | None = None) -> None:
        """Decrease current value by step size, or provided value."""
        self.value = self.get_value() - (val if val is not None else self.step)

    # overriding because at least some backends don't have a step value for ProgressBar
    @property  # type: ignore
    def step(self) -> float:
        """Step size for widget values."""
        return self._step

    @step.setter
    def step(self, value: float) -> None:
        self._step = value


@backend_widget
class Slider(SliderWidget[int]):
    """A slider widget to adjust an integer value within a range."""


@backend_widget
class FloatSlider(SliderWidget[float]):
    """A slider widget to adjust an integer value within a range."""


@backend_widget
class RangeSlider(MultiValuedSliderWidget):
    """A slider widget to adjust a range between two integer values within a range."""


@backend_widget
class FloatRangeSlider(MultiValuedSliderWidget):
    """A slider widget to adjust a range defined by two float values within a range."""


@merge_super_sigs
class LogSlider(TransformedRangedWidget):
    """A slider widget to adjust a numerical value logarithmically within a range.

    Parameters
    ----------
    base : Enum, Iterable, or Callable
        The base to use for the log, by default math.e.
    """

    _widget: protocols.SliderWidgetProtocol

    def __init__(
        self,
        min: float = 1,
        max: float = 100,
        base: float = math.e,
        tracking: bool = True,
        **kwargs: Unpack[WidgetKwargs],
    ):
        self._base = base
        app = use_app()
        assert app.native
        kwargs["widget_type"] = app.get_obj("Slider")
        super().__init__(min=min, max=max, **kwargs)
        self.tracking = tracking

    @property
    def tracking(self) -> bool:
        """Return whether slider tracking is enabled.

        If tracking is enabled (the default), the slider emits the changed()
        signal while the slider is being dragged. If tracking is disabled,
        the slider emits the changed() signal only when the user releases
        the slider.
        """
        return self._widget._mgui_get_tracking()

    @tracking.setter
    def tracking(self, value: bool) -> None:
        """Set whether slider tracking is enabled."""
        self._widget._mgui_set_tracking(value)

    @property
    def _scale(self) -> float:
        minv = math.log(self.min, self.base)
        maxv = math.log(self.max, self.base)
        return (maxv - minv) / (self._max_pos - self._min_pos)

    def _value_from_position(self, position: float) -> float:
        minv = math.log(self.min, self.base)
        return math.pow(self.base, minv + self._scale * (position - self._min_pos))

    def _position_from_value(self, value: float) -> float:
        minv = math.log(self.min, self.base)
        pos = (math.log(value, self.base) - minv) / self._scale + self._min_pos
        return int(pos)

    @property
    def base(self) -> float:
        """Return base used for the log."""
        return self._base

    @base.setter
    def base(self, base: float) -> None:
        prev = self.value
        self._base = base
        self.value = prev


@backend_widget
class ComboBox(CategoricalWidget):
    """A dropdown menu, allowing selection between multiple choices."""


@backend_widget
class Select(CategoricalWidget):
    """A list of options, allowing selection between multiple choices."""

    _allow_multiple = True


@merge_super_sigs
class RadioButtons(CategoricalWidget, _OrientationMixin):  # type: ignore
    """An exclusive group of radio buttons, providing a choice from multiple choices."""

    def __init__(
        self,
        choices: ChoicesType = (),
        orientation: str = "vertical",
        **kwargs: Unpack[WidgetKwargs],
    ) -> None:
        app = use_app()
        assert app.native
        kwargs["widget_type"] = app.get_obj("RadioButtons")
        super().__init__(choices=choices, **kwargs)
        self.orientation = orientation


@backend_widget
class Container(ContainerWidget[WidgetVar]):
    """A Widget to contain other widgets.

    Note that `Container` implements the
    [`typing.MutableSequence`][typing.MutableSequence]
    interface, so you can use it like a list to add and remove widgets.
    """


@backend_widget
class Dialog(DialogWidget):
    """A modal container."""


@backend_widget
class MainWindow(MainWindowWidget):
    """A Widget to contain other widgets, includes a menu bar."""


@merge_super_sigs
class FileEdit(Container):
    """A LineEdit widget with a button that opens a FileDialog.

    Parameters
    ----------
    mode : FileDialogMode or str
        - ``'r'`` returns one existing file.
        - ``'rm'`` return one or more existing files.
        - ``'w'`` return one file name that does not have to exist.
        - ``'d'`` returns one existing directory.
    filter : str, optional
        The filter is used to specify the kind of files that should be shown.
        It should be a glob-style string, like ``'*.png'`` (this may be
        backend-specific)
    """

    def __init__(
        self,
        mode: FileDialogMode = FileDialogMode.EXISTING_FILE,
        filter: str | None = None,
        nullable: bool = False,
        **kwargs: Unpack[ContainerKwargs],
    ) -> None:
        # use empty string as a null value
        value = kwargs.pop("value", None)  # type: ignore [typeddict-item]
        if value is None:
            value = ""
        self.line_edit = LineEdit(value=value)
        self.choose_btn = PushButton()
        self.mode = mode  # sets the button text too
        self.filter = filter
        self._nullable = nullable
        kwargs["widgets"] = [self.line_edit, self.choose_btn]
        kwargs["labels"] = False
        kwargs["layout"] = "horizontal"
        super().__init__(**kwargs)
        self.margins = (0, 0, 0, 0)
        self._show_file_dialog = use_app().get_obj("show_file_dialog")
        self.choose_btn.changed.disconnect()
        self.line_edit.changed.disconnect()
        self.choose_btn.changed.connect(self._on_choose_clicked)
        self.line_edit.changed.connect(lambda: self.changed.emit(self.value))

    @property
    def mode(self) -> FileDialogMode:
        """Mode for the FileDialog."""
        return self._mode

    @mode.setter
    def mode(self, value: FileDialogMode | str) -> None:
        self._mode = FileDialogMode(value)
        self.choose_btn.text = self._btn_text

    @property
    def _btn_text(self) -> str:
        if self.mode is FileDialogMode.EXISTING_DIRECTORY:
            return "Choose directory"
        else:
            return "Select file" + ("s" if self.mode.name.endswith("S") else "")

    def _on_choose_clicked(self) -> None:
        _p = self.value
        if _p:
            start_path: Path = _p[0] if isinstance(_p, tuple) else _p
            _start_path: str | None = os.fspath(start_path.expanduser().absolute())
        else:
            _start_path = None
        result = self._show_file_dialog(
            self.mode,
            caption=self._btn_text,
            start_path=_start_path,
            filter=self.filter,
        )
        if result:
            self.value = result

    @property
    def value(self) -> tuple[Path, ...] | Path | None:
        """Return current value of the widget.  This may be interpreted by backends."""
        text = self.line_edit.value
        if self._nullable and not text:
            return None
        if self.mode is FileDialogMode.EXISTING_FILES:
            return tuple(Path(p) for p in text.split(", ") if p.strip())
        return Path(text)

    @value.setter
    def value(self, value: Sequence[PathLike] | PathLike | None) -> None:
        """Set current file path."""
        if value is None and self._nullable:
            value = ""
        elif isinstance(value, (list, tuple)):
            value = ", ".join(os.fspath(Path(p).expanduser().absolute()) for p in value)
        elif isinstance(value, (str, Path)):
            value = os.fspath(Path(value).expanduser().absolute())
        else:
            raise TypeError(
                f"value must be a string, or list/tuple of strings, got {type(value)}"
            )
        self.line_edit.value = value

    def __repr__(self) -> str:
        """Return string representation."""
        return f"FileEdit(mode={self.mode.value!r}, value={self.value!r})"


@merge_super_sigs
class RangeEdit(Container[SpinBox]):
    """A widget to represent a python range object, with start/stop/step.

    A range object produces a sequence of integers from start (inclusive)
    to stop (exclusive) by step.  range(i, j) produces i, i+1, i+2, ..., j-1.
    start defaults to 0, and stop is omitted!  range(4) produces 0, 1, 2, 3.
    These are exactly the valid indices for a list of 4 elements.
    When step is given, it specifies the increment (or decrement).

    Parameters
    ----------
    start : int, optional
        The range start value, by default 0
    stop : int, optional
        The range stop value, by default 10
    step : int, optional
        The range step value, by default 1
    """

    def __init__(
        self,
        start: int = 0,
        stop: int = 10,
        step: int = 1,
        min: int | tuple[int, int, int] | None = None,
        max: int | tuple[int, int, int] | None = None,
        **kwargs: Unpack[ContainerKwargs],
    ) -> None:
        value = kwargs.pop("value", None)  # type: ignore [typeddict-item]
        if value is not None and value is not Undefined:
            if not all(hasattr(value, x) for x in ("start", "stop", "step")):
                raise TypeError(f"Invalid value type for {type(self)}: {type(value)}")
            start, stop, step = value.start, value.stop, value.step
        minstart, minstop, minstep = self._validate_min_max(min, "min", -9999999)
        maxstart, maxstop, maxstep = self._validate_min_max(max, "max", 9999999)
        self.start = SpinBox(value=start, min=minstart, max=maxstart, name="start")
        self.stop = SpinBox(value=stop, min=minstop, max=maxstop, name="stop")
        self.step = SpinBox(value=step, min=minstep, max=maxstep, name="step")
        kwargs["widgets"] = [self.start, self.stop, self.step]
        kwargs.setdefault("layout", "horizontal")
        kwargs.setdefault("labels", True)
        kwargs.pop("nullable", None)  # type: ignore [typeddict-item]
        super().__init__(**kwargs)

    @classmethod
    def _validate_min_max(
        cls, arg: int | tuple[int, int, int] | None, name: str, default: int
    ) -> tuple[int, int, int]:
        """Validate input to the min/max arguments."""
        if isinstance(arg, (int, float)):
            return (int(arg),) * 3
        elif isinstance(arg, (list, tuple)):
            if len(arg) != 3:
                raise ValueError(f"{name} sequence must be length 3")
            return cast("tuple[int, int, int]", tuple(int(x) for x in arg))
        elif arg is not None:
            raise TypeError("min must be an integer or a 3-tuple of integers")
        else:
            return (int(default),) * 3

    @property
    def value(self) -> range:
        """Return current value of the widget.  This may be interpreted by backends."""
        return range(self.start.value, self.stop.value, self.step.value)

    @value.setter
    def value(self, value: range) -> None:
        """Set current file path."""
        self.start.value = value.start
        self.stop.value = value.stop
        self.step.value = value.step

    def __repr__(self) -> str:
        """Return string representation."""
        return f"<{self.__class__.__name__} value={self.value!r}>"


class SliceEdit(RangeEdit):
    """A widget to represent `slice` objects, with start/stop/step.

    slice(stop)
    slice(start, stop[, step])

    Slice objects may be used for extended slicing (e.g. a[0:10:2])
    """

    @property  # type: ignore
    def value(self) -> slice:
        """Return current value of the widget.  This may be interpreted by backends."""
        return slice(self.start.value, self.stop.value, self.step.value)

    @value.setter
    def value(self, value: slice) -> None:
        """Set current file path."""
        self.start.value = value.start
        self.stop.value = value.stop
        self.step.value = value.step


class _ListEditChildWidget(Container[Widget]):
    """A widget to represent a single element of a ListEdit widget."""

    def __init__(self, widget: ValueWidget):
        btn = PushButton(text="-")
        super().__init__(widgets=[widget, btn], layout="horizontal", labels=False)
        self.btn_minus = btn
        self.value_widget = widget
        self.margins = (0, 0, 0, 0)

        btn.changed.disconnect()
        widget.changed.disconnect()
        widget.changed.connect(self.changed.emit)
        btn.max_height = btn.max_width = use_app().get_obj("get_text_width")("-") + 4

    @property
    def value(self) -> Any:
        """Return value of the child widget."""
        return self.value_widget.value

    @value.setter
    def value(self, value: Any) -> None:
        """Set value of the child widget."""
        self.value_widget.value = value

@merge_super_sigs
class ListEdit(Container[ValueWidget[_V]]):
    """A widget to represent a list of values.

    A ListEdit container can create a list with multiple objects of same type. It
    will contain many child widgets and their value is represented as a Python list
    object. If a list is given as the initial value, types of child widgets are
    determined from the contents. Number of contents can be adjusted with +/-
    buttons.

    Parameters
    ----------
    value : Iterable, optional
        The starting value for the widget.
    nullable : bool
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    options: dict, optional
        Widget options of child widgets.
    """

    def __init__(
        self,
        value: Iterable[_V] | _Undefined = Undefined,
        nullable: bool = False,
        options: dict | None = None,
        **container_kwargs: Unpack[ContainerKwargs],
    ) -> None:
        self._args_type: type | None = None
        self._nullable = nullable
        container_kwargs.setdefault("layout", "horizontal")
        container_kwargs.setdefault("labels", False)
        super().__init__(**container_kwargs)
        self.margins = (0, 0, 0, 0)

        if not isinstance(value, _Undefined):
            # check type consistency
            types = {type(a) for a in value}
            if len(types) > 1:
                raise TypeError("values have inconsistent types.")
            if self._args_type is None:
                self._args_type = types.pop()
            _value: Iterable[_V] = value
        else:
            _value = []

        self._child_options = options or {}

        button_plus = PushButton(text="+", name="plus")

        self.append(button_plus)  # type: ignore
        button_plus.changed.disconnect()
        button_plus.changed.connect(lambda: self._append_value())

        for a in _value:
            self._append_value(a)

        self.btn_plus = button_plus

    @property
    def annotation(self) -> Any:
        """Return type annotation for the parameter represented by the widget.

        ForwardRefs will be resolve when setting the annotation. For ListEdit,
        annotation will be like 'list[str]'.
        """
        return self._annotation

    @annotation.setter
    def annotation(self, value: Any) -> None:
        if value is None:
            self._annotation = None
            self._args_type = None
            return

        value_resolved = resolve_single_type(value)
        if isinstance(value, (str, ForwardRef)):
            value = value_resolved
        # unwrap annotated (options are not needed to normalize `annotation`)
        while get_origin(value) is Annotated:
            value = get_args(value)[0]
        arg: type | None = None

        if value_resolved and value_resolved is not inspect.Parameter.empty:
            orig = get_origin(value_resolved) or value_resolved
            if not (safe_issubclass(orig, list) or isinstance(orig, list)):
                raise TypeError(
                    f"cannot set annotation {value} to {type(self).__name__}."
                )
            args = get_args(value)
            arg = args[0] if len(args) > 0 else None
            args_resolved = get_args(value_resolved)
            if len(args_resolved) > 0:
                value = List[args_resolved[0]]  # type: ignore
            else:
                value = list

        self._annotation = value
        self._args_type = arg

    def __delitem__(self, key: int | slice) -> None:
        """Delete child widget(s)."""
        super().__delitem__(key)
        self.changed.emit(self.value)

    def _append_value(self, value: _V | _Undefined = Undefined) -> None:
        """Create a new child value widget and append it."""
        i = len(self) - 1

        _value_widget = create_widget(
            annotation=self._args_type,
            name=f"value_{i}",
            options=self._child_options,
        )
        widget = _ListEditChildWidget(cast(ValueWidget, _value_widget))
        # connect the minus-button-clicked event
        widget.btn_minus.changed.connect(lambda: self.remove(widget))

        # _ListEditChildWidget is technically a ValueWidget.
        self.insert(i, widget)  # type: ignore

        widget.changed.disconnect()

        # Value must be set after new widget is inserted because it could be
        # valid only after same parent is shared between widgets.
        if value is Undefined and i > 0:
            # copy value from the previous child widget if possible
            value = self[i - 1].value
        if value is not Undefined:
            widget.value = value

        widget.changed.connect(lambda: self.changed.emit(self.value))
        self.changed.emit(self.value)

    @property
    def value(self) -> list[_V]:
        """Return current value as a list object."""
        return list(ListDataView(self))

    @value.setter
    def value(self, vals: Iterable[_V]) -> None:
        with self.changed.blocked():
            del self[:-1]
            for v in vals:
                self._append_value(v)
        self.changed.emit(self.value)

    @property
    def data(self) -> ListDataView[_V]:
        """Return a data view of current value."""
        return ListDataView(self)

    @data.setter
    def data(self, vals: Iterable[_V]) -> None:
        self.value = vals  # type: ignore[assignment]


class ListDataView(Generic[_V]):
    """Data view of ListEdit."""

    def __init__(self, obj: ListEdit[_V]):
        self._obj = obj
        self._widgets = list(obj[:-1])

    def __repr__(self) -> str:
        """Return list-like representation."""
        return f"{self.__class__.__name__}({list(self)!r})"

    def __len__(self) -> int:
        """Length as a list."""
        return len(self._widgets)

    def __eq__(self, other: object) -> bool:
        """Compare as a list."""
        return list(self) == other

    @overload
    def __getitem__(self, i: int) -> _V: ...

    @overload
    def __getitem__(self, key: slice) -> list[_V]: ...

    def __getitem__(self, key: int | slice) -> _V | list[_V]:
        """Slice as a list."""
        if isinstance(key, int):
            return self._widgets[key].value
        elif isinstance(key, slice):
            return [w.value for w in self._widgets[key]]
        else:
            raise TypeError(
                f"list indices must be integers or slices, not {type(key).__name__}"
            )

    @overload
    def __setitem__(self, key: int, value: _V) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: _V | Iterable[_V]) -> None: ...

    def __setitem__(self, key: int | slice, value: _V | Iterable[_V]) -> None:
        """Update widget value."""
        if isinstance(key, int):
            self._widgets[key].value = cast(_V, value)
        elif isinstance(key, slice):
            with self._obj.changed.blocked():
                if isinstance(value, type(self._widgets[0].value)):
                    for w in self._widgets[key]:
                        w.value = value
                else:
                    for w, v in zip(self._widgets[key], value):  # type: ignore
                        w.value = v
            self._obj.changed.emit(self._obj.value)
        else:
            raise TypeError(
                f"list indices must be integers or slices, not {type(key).__name__}"
            )

    @overload
    def __delitem__(self, key: int) -> None: ...

    @overload
    def __delitem__(self, key: slice) -> None: ...

    def __delitem__(self, key: int | slice) -> None:
        """Delete widget at the key(s)."""
        self._obj.__delitem__(key)

    def __iter__(self) -> Iterator[_V]:
        """Iterate over values of child widgets."""
        for w in self._widgets:
            yield w.value


@merge_super_sigs
class TupleEdit(Container[ValueWidget]):
    """A widget to represent a tuple of values.

    A TupleEdit container has several child widgets of different type. Their value is
    represented as a Python tuple object. If a tuple is given as the initial value,
    types of child widgets are determined one by one. Unlike ListEdit, number of
    contents is not editable.

    Parameters
    ----------
    value : Iterable, optional
        The starting value for the widget.
    nullable : bool
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    options: dict, optional
        Widget options of child widgets.
    """

    def __init__(
        self,
        value: Iterable[_V] | _Undefined = Undefined,
        *,
        nullable: bool = False,
        options: dict | None = None,
        **container_kwargs: Unpack[ContainerKwargs[ValueWidget]],
    ) -> None:
        self._nullable = nullable
        self._args_types: tuple[type, ...] | None = None
        container_kwargs.setdefault("labels", False)
        container_kwargs.setdefault("layout", "horizontal")
        super().__init__(**container_kwargs)
        self._child_options = options or {}
        self.margins = (0, 0, 0, 0)

        if not isinstance(value, _Undefined):
            if self._args_types is None:
                self._args_types = tuple(type(a) for a in value)
            _value: Iterable[Any] = value
        elif self._args_types is not None:
            _value = (Undefined,) * len(self._args_types)
        else:
            raise ValueError(
                "Either 'value' or 'annotation' must be specified in "
                f"{type(self).__name__}."
            )

        for a in _value:
            i = len(self)
            widget = cast(
                ValueWidget,
                create_widget(
                    value=a,
                    annotation=self._args_types[i],
                    name=f"value_{i}",
                    options=self._child_options,
                ),
            )
            self.insert(i, widget)
            widget.changed.disconnect()
            widget.changed.connect(lambda: self.changed.emit(self.value))

    @property
    def annotation(self) -> Any:
        """Return type annotation for the parameter represented by the widget.

        ForwardRefs will be resolve when setting the annotation. For TupleEdit,
        annotation will be like 'tuple[str, int]'.
        """
        return self._annotation

    @annotation.setter
    def annotation(self, value: Any) -> None:
        if value is None:
            self._annotation = None
            self._args_types = None
            return

        value_resolved = resolve_single_type(value)
        if isinstance(value, (str, ForwardRef)):
            value = value_resolved
        # unwrap annotated (options are not needed to normalize `annotation`)
        while get_origin(value) is Annotated:
            value = get_args(value)[0]
        args: tuple[type, ...] | None = None

        if value_resolved and value_resolved is not inspect.Parameter.empty:
            orig = get_origin(value_resolved)
            if not (safe_issubclass(orig, tuple) or isinstance(orig, tuple)):
                raise TypeError(
                    f"cannot set annotation {value} to {type(self).__name__}."
                )
            args = get_args(value)
            args_resolved = get_args(value_resolved)
            value = Tuple[args_resolved]

        self._annotation = value
        self._args_types = args

    @property
    def value(self) -> tuple:
        """Return current value as a tuple."""
        return tuple(w.value for w in self)

    @value.setter
    def value(self, vals: Sequence) -> None:
        if len(vals) != len(self):
            raise ValueError("Length of tuple does not match.")

        with self.changed.blocked():
            for w, v in zip(self, vals):
                w.value = v
        self.changed.emit(self.value)


@backend_widget
class ToolBar(ToolBarWidget):
    """Toolbar that contains a set of controls."""


class _LabeledWidget(Container):
    """Simple container that wraps a widget and provides a label."""

    def __init__(  # type: ignore [misc]  # overlap between argument names
        self,
        widget: Widget,
        label: str | None = None,
        position: str = "left",
        **kwargs: Unpack[ContainerKwargs],
    ) -> None:
        kwargs["layout"] = "horizontal" if position in {"left", "right"} else "vertical"
        self._inner_widget = widget
        widget._labeled_widget_ref = ref(self)
        _visible = False if widget._explicitly_hidden else None
        self._label_widget = Label(value=label or widget.label, tooltip=widget.tooltip)
        super().__init__(**kwargs, visible=_visible)
        self.native_parent_changed.disconnect()  # don't need _LabeledWidget to trigger
        self.labels = False  # important to avoid infinite recursion during insert!
        self._inner_widget.label_changed.connect(self._on_label_change)
        for w in [self._label_widget, widget]:
            with w.native_parent_changed.blocked():
                self.append(w)
        self.margins = (0, 0, 0, 0)

    def __repr__(self) -> str:
        """Return string representation."""
        return f"Labeled{self._inner_widget!r}"

    @property
    def value(self) -> Any:
        return getattr(self._inner_widget, "value", None)

    @value.setter
    def value(self, value: Any) -> None:
        if hasattr(self._inner_widget, "value"):
            self._inner_widget.value = value

    @property
    def label(self) -> str:
        return self._label_widget.label

    @label.setter
    def label(self, label: str) -> None:
        self._label_widget.label = label

    def _on_label_change(self, value: str) -> None:
        self._label_widget.value = value

    @property
    def label_width(self) -> int:
        return self._label_widget.width

    @label_width.setter
    def label_width(self, width: int) -> None:
        self._label_widget.min_width = width


@backend_widget
class QuantityEdit(ValueWidget):
    """A combined `LineEdit` and `ComboBox` to edit a `pint.Quantity`."""



"""# Quantities with pint

Pint is a Python package to define, operate and manipulate physical quantities:
the product of a numerical value and a unit of measurement.
It allows arithmetic operations between them and conversions
from and to different units.
https://pint.readthedocs.io/en/stable/
"""

from pint import Quantity

from magicgui import magicgui


@magicgui
def widget(q=Quantity("1 ms")):
    """Widget allowing users to input quantity measurements."""
    print(q)


widget.show(run=True)



"""# Dropdown selection widget

Choices for dropdowns can be provided in a few different ways.
"""

from enum import Enum

from magicgui import magicgui, widgets


class Medium(Enum):
    """Enum for various media and their refractive indices."""

    Oil = 1.515
    Water = 1.333
    Air = 1.0003


@magicgui(ri={"choices": ["Oil", "Water", "Air"]}, auto_call=True)
def as_list(ri="Water"):
    """Function decorated with magicgui list of choices."""
    print("refractive index is", Medium[ri].value)


@magicgui(auto_call=True)
def as_enum(ri: Medium = Medium.Water):
    """Function decorated with magicgui and enumeration."""
    print("refractive index is", ri.value)


@magicgui(
    ri={"choices": [("Oil", 1.515), ("Water", 1.33), ("Air", 1.0)]}, auto_call=True
)
def as_2tuple(ri=1.33):
    """Function decorated with magicgui tuple of choices."""
    print("refractive index is", ri)


def get_choices(gui):
    """Function returning tuple of material and refractive index value."""
    return [("Oil", 1.515), ("Water", 1.33), ("Air", 1.0)]


@magicgui(ri={"choices": get_choices}, auto_call=True)
def as_function(ri: float):
    """Function to calculate refractive index."""
    print("refractive index is", ri)


container = widgets.Container(
    widgets=[as_list, as_enum, as_2tuple, as_function], layout="vertical"
)
container.show(run=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Tuple, TypeVar, Union

from ._widget import Widget

if TYPE_CHECKING:
    from magicgui.widgets import protocols

T = TypeVar("T", int, float, Tuple[Union[int, float], ...])
DEFAULT_MIN = 0.0
DEFAULT_MAX = 1000.0


class ToolBarWidget(Widget):
    """Widget with a value, Wraps ValueWidgetProtocol.

    Parameters
    ----------
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    _widget: protocols.ToolBarProtocol

    def __init__(self, **base_widget_kwargs: Any) -> None:
        super().__init__(**base_widget_kwargs)

    def add_button(
        self, text: str = "", icon: str = "", callback: Callable | None = None
    ) -> None:
        """Add an action to the toolbar."""
        self._widget._mgui_add_button(text, icon, callback)

    def add_separator(self) -> None:
        """Add a separator line to the toolbar."""
        self._widget._mgui_add_separator()

    def add_spacer(self) -> None:
        """Add a spacer to the toolbar."""
        self._widget._mgui_add_spacer()

    def add_widget(self, widget: Widget) -> None:
        """Add a widget to the toolbar."""
        self._widget._mgui_add_widget(widget)

    @property
    def icon_size(self) -> tuple[int, int] | None:
        """Return the icon size of the toolbar."""
        return self._widget._mgui_get_icon_size()

    @icon_size.setter
    def icon_size(self, size: int | tuple[int, int] | None) -> None:
        """Set the icon size of the toolbar."""
        self._widget._mgui_set_icon_size(size)

    def clear(self) -> None:
        """Clear the toolbar."""
        self._widget._mgui_clear()
